from __future__ import division

import numpy as np
from .coordinates import RegularCoords, SeparatedCoords, UnstructuredCoords
from .field import Field
from .cartesian_grid import CartesianGrid
from .grid import Grid

def make_uniform_grid(dims, extent, center=0, has_center=False):
	'''Create a uniformly-spaced :class:`Grid` of a certain shape and size.

	Parameters
	----------
	dims : scalar or ndarray
		The number of points in each dimension. If this is a scalar, it will
		be multiplexed over all dimensions.
	extent : scalar or ndarray
		The total extent of the grid in each dimension.
	center : scalar or ndarray
		The center point. The grid will by symmetric around this point.
	has_center : boolean
		Does the grid has to have the center as one of its points. If this is
		False, this does not mean that the grid will not have the center.

	Returns
	-------
	Grid
		A :class:`Grid` with :class:`RegularCoords`.
	'''
	num_dims = max(np.array([dims]).shape[-1], np.array([extent]).shape[-1], np.array([center]).shape[-1])

	dims = (np.ones(num_dims) * dims).astype('int')
	extent = (np.ones(num_dims) * extent).astype('float')
	center = (np.ones(num_dims) * center).astype('float')

	delta = extent / dims
	zero = -extent / 2 + center + delta / 2

	if has_center:
		zero -= delta / 2 * (1 - np.mod(dims, 2))

	return CartesianGrid(RegularCoords(delta, dims, zero))

def make_pupil_grid(dims, diameter=1):
	'''Makes a new :class:`Grid`, meant for descretisation of a pupil-plane wavefront.

	This grid is symmetric around the origin, and therefore has no point exactly on
	the origin for an even number of pixels.

	Parameters
	----------
	dims : ndarray or integer
		The number of pixels per dimension. If this is an integer, this number
		of pixels is used for all dimensions.
	diameter : ndarray or scalar
		The diameter of the grid in each dimension. If this is a scalar, this diameter
		is used for all dimensions.

	Returns
	-------
	Grid
		A :class:`CartesianGrid` with :class:`RegularCoords`.
	'''
	diameter = (np.ones(2) * diameter).astype('float')
	return make_uniform_grid(dims, diameter)

def make_focal_grid_from_pupil_grid(pupil_grid, q=1, num_airy=None, focal_length=1, wavelength=1):
	'''Make a grid for a focal plane from a pupil grid.

	Calculate the focal grid corresponding to the pupil grid, using an FFT grid as a guide. The
	resulting grid focal will always contain the origin (0, 0) point. The extent of the pupil
	grid will be used as the diameter of the pupil. If the pupil is undersized on the pupil
	grid, the resulting focal grid needs to be rescaled manually.

	.. note::
		In almost all cases, it is preferable to use :func:`make_focal_grid`. This
		function allows you to directly set the diameter, and doesn't require the user to pass
		the pupil grid as an argument. `make_focal_grid_from_pupil_grid()` retains old functionality
		and serves as a backwards compatibility function, due to its ubiquitous usage in HCIPy code.

	Parameters
	----------
	pupil_grid : Grid
		The pupil grid for which the focal grid needs to be calculated. The extent of this Grid
		is used as the diameter of the pupil.
	q : scalar or array_like
		The number of pixels per resolution element (= lambda f / D).
	num_airy : scalar or array_like
		The spatial extent of the grid in radius in resolution elements (= lambda f / D).
	focal_length : scalar
		The focal length used for calculating the spatial resolution at the focal plane.
	wavelength : scalar
		The reference wavelength used for calculating the spatial resolution at the focal plane.

	Returns
	-------
	Grid
		A Grid describing the sampling for a focal plane.
	'''
	from ..fourier import make_fft_grid

	f_lambda = focal_length * wavelength
	if num_airy is None:
		fov = 1
	else:
		fov = (num_airy * np.ones(pupil_grid.ndim, dtype='float')) / (pupil_grid.shape / 2)

	if np.max(fov) > 1:
		import warnings
		warnings.warn('Focal grid is larger than the maximum allowed angle (fov=%.03f). You may see wrapping when doing propagations.' % np.max(fov), stacklevel=2)

	uv = make_fft_grid(pupil_grid, q, fov)
	focal_grid = uv.scaled(f_lambda / (2*np.pi))

	return focal_grid

def make_focal_grid(q, num_airy, spatial_resolution=None, pupil_diameter=None, focal_length=None, f_number=None, reference_wavelength=None):
	r'''Make a grid for a focal plane.

	This grid will be a CartesianGrid with RegularCoords, and supports different resolutions, samplings,
	or extent in x and y. If `spatial_resolution` is 1, then the grid will be returned in normalized
	(ie. homogenized) coordinates. Otherwise, it will be in physical units. The spatial resolution is
	defined by:

	.. math:: \Delta x = \lambda f / D = \lambda F

	where :math:`\lambda` is the wavelength, :math:`f` is the effective focal length before the focal plane,
	:math:`D` is the diameter of the pupil, and :math:`F` is the F-number of the incoming light beam.
	You can supply either the spatial resolution or a set of focal length, reference wavelength and pupil diameter,
	or a set of F-number and reference wavelength. If none are supplied, a spatial resolution of 1 will be
	assumed, meaning normalized units.

	The grid will always contain the origin (0, 0) point.

	Parameters
	----------
	q : scalar or array_like
		The number of pixels per resolution element (= lambda f / D).
	num_airy : scalar or array_like
		The spatial extent of the grid in radius in resolution elements (= lambda f / D).
	spatial_resolution : scalar	or array_like
		The physical size of a resolution element (= lambda f / D). It this is not given,
		the spatial resolution will be calculated from the given `focal_length`,
		`reference_wavelength` and `pupil_diameter`.
	pupil_diameter : scalar or array_like
		The diameter of the pupil. If it is an array, this indicates the diameter in x and y.
	focal_length : scalar
		The focal length used for calculating the spatial resolution at the focal plane.
	f_number : scalar or array_like
		The F number, also known as focal ratio, at the focal plane. If this is given, it overrides
		the given pupil diameter and focal length.
	reference_wavelength : scalar
		The reference wavelength used for calculating the spatial resolution at the focal plane.

	Returns
	-------
	Grid
		A Grid describing the sampling for a focal plane.

	Raises
	------
	ValueError
		If both no spatial resolution and no complete set of (focal length, reference wavelength and pupil diameter) was supplied.
	'''
	if isinstance(q, Grid):
		raise ValueError('The function signature was changed as of HCIPy 0.3.0. Please use the new signature (prefered), or use make_focal_grid_from_pupil_grid() if you want to retain old behaviour.')

	if spatial_resolution is None:
		if f_number is None:
			if pupil_diameter is None or focal_length is None:
				if reference_wavelength is None:
					spatial_resolution = 1
				else:
					raise ValueError('You only supplied a reference wavelength and forgot to supply either an f_number or a (pupil_diameter, focal_length).')
			else:
				f_number = focal_length / pupil_diameter

		if spatial_resolution is None:
			if reference_wavelength is None:
				raise ValueError('You supplied an f_number or (pupil_diameter, focal_length), and forgot to supply a reference wavelength.')
			else:
				spatial_resolution = f_number * reference_wavelength

	delta = spatial_resolution / q * np.ones(2)
	dims = (2 * num_airy * q * np.ones(2)).astype('int')
	zero = delta * (-dims / 2 + np.mod(dims, 2) * 0.5)

	return CartesianGrid(RegularCoords(delta, dims, zero))

def make_hexagonal_grid(circum_diameter, n_rings, pointy_top=False, center=None):
	'''Make a regular hexagonal grid.

	Parameters
	----------
	circum_diameter : scalar
		The circum diameter of the hexagons in the grid.
	n_rings : integer
		The number of rings in the grid.
	pointy_top : boolean
		If the hexagons contained in the grid.
	center : ndarray
		The center of the grid in cartesian coordinates.

	Returns
	-------
	Grid
		A :class:`CartesianGrid` with `UnstructuredCoords`, indicating the
		center of the hexagons.
	'''
	if center is None:
		center = np.zeros(2)

	apothem = circum_diameter * np.sqrt(3) / 4

	q = [0]
	r = [0]

	for n in range(1, n_rings + 1):
		#top
		q += list(range(n, 0, -1))
		r += list(range(0, n))
		# right top
		q += list(range(0, -n, -1))
		r += [n] * n
		# right bottom
		q += [-n] * n
		r += list(range(n, 0, -1))
		# bottom
		q += list(range(-n, 0))
		r += list(range(0, -n, -1))
		# left bottom
		q += list(range(0, n))
		r += [-n] * n
		# left top
		q += [n] * n
		r += list(range(-n, 0))

	x = (-np.array(q) + np.array(r)) * circum_diameter / 2 + center[0]
	y = (np.array(q) + np.array(r)) * apothem * 2 + center[1]

	weight = 2 * apothem**2 * np.sqrt(3)

	if pointy_top:
		return CartesianGrid(UnstructuredCoords((x, y)), weight)
	else:
		return CartesianGrid(UnstructuredCoords((y, x)), weight)

def make_chebyshev_grid(dims, minimum=None, maximum=None):
	if minimum is None:
		minimum = -1

	if maximum is None:
		maximum = 1

	dims = np.array(dims)
	minimum = np.ones(len(dims)) * minimum
	maximum = np.ones(len(dims)) * maximum

	middles = (minimum + maximum) / 2
	intervals = (maximum - minimum) / 2

	sep_coords = []
	for dim, middle, interval in zip(dims, middles, intervals):
		c = np.cos(np.pi * (2 * np.arange(dim) + 1) / (2.0 * dim))
		c = middle + interval * c
		sep_coords.append(c)

	return CartesianGrid(SeparatedCoords(sep_coords))

def make_supersampled_grid(grid, oversampling):
	'''Make a new grid that oversamples by a factor `oversampling`.

	.. note ::
		The Grid `grid` must be a grid with separable coordinates.

	Parameters
	----------
	grid : Grid
		The grid that we want to oversample.
	oversampling : integer or scalar or ndarray
		The factor by which to oversample. If this is a scalar, it will be rounded to
		the nearest integer. If this is an array, a different oversampling factor will
		be used for each dimension.

	Returns
	-------
	Grid
		The oversampled grid.
	'''
	oversampling = (np.round(oversampling)).astype('int')

	if grid.is_regular:
		delta_new = grid.delta / oversampling
		zero_new = grid.zero - grid.delta / 2 + delta_new / 2
		dims_new = grid.dims * oversampling

		return grid.__class__(RegularCoords(delta_new, dims_new, zero_new))
	elif grid.is_separated:
		raise NotImplementedError()

	raise ValueError('Cannot create a supersampled grid from a non-separated grid.')

def make_subsampled_grid(grid, undersampling):
	'''Make a new grid that undersamples by a factor `undersampling`.

	.. note ::
		The dimensions of the `grid` must be divisible by `undersampling`.

	Parameters
	----------
	grid : Grid
		The grid that we want to oversample.
	undersampling : integer or scalar or ndarray
		The factor by which to undersample. If this is a scalar, it will be rounded to
		the nearest integer. If this is an array, a different undersampling factor will
		be used for each dimension.

	Returns
	-------
	Grid
		The undersampled grid.
	'''
	undersampling = (np.round(undersampling)).astype('int')

	if grid.is_regular:
		delta_new = grid.delta * undersampling
		zero_new = grid.zero - grid.delta / 2 + delta_new / 2
		dims_new = grid.dims // undersampling

		return grid.__class__(RegularCoords(delta_new, dims_new, zero_new))
	elif grid.is_separated:
		raise NotImplementedError()

	raise ValueError("Cannot create a subsampled grid from a non-separated grid.")

def subsample_field(field, subsampling, new_grid=None, statistic='mean'):
	'''Average the field over subsampling pixels in each dimension.

	.. note ::
		The dimensions of the grid of `field` must be divisible by `subsampling`.

	Parameters
	----------
	field : Field
		The field to subsample. The grid of this field must have the right
		dimensions to be able to be subsampled.
	subsampling : integer or scalar or ndarray
		The subsampling factor. If this is a scalar, it will be rounded to the
		nearest integer. If this is an array, the subsampling factor will be
		different for each dimension.
	new_grid : Grid
		If this grid is given, no new grid will be calculated and this grid will
		be used instead. This saves on calculation time if your new grid is already
		known beforehand.
	statistic : string
		The statistic to compute (default is 'mean').
		The following statistics are available:
		* 'mean': compute the mean of values for points within each superpixel.
		* 'sum': compute the sum of values for points within each superpixel. This is identical to a weighted histogram.
		* 'min': compute the minimum of values for points within each superpixel.
		* 'max': compute the maximum of values for point within each superpixel.

	Returns
	-------
	Field
		The subsampled field.
	'''
	subsampling = (np.round(subsampling)).astype('int')

	if new_grid is None:
		new_grid = make_subsampled_grid(field.grid, subsampling)

	reshape = []
	axes = []
	for i, s in enumerate(new_grid.shape):
		reshape.extend([s, subsampling])
		axes.append(2 * i + 1)

	if field.tensor_order > 0:
		reshape = list(field.tensor_shape) + reshape
		axes = np.array(axes) + field.tensor_order
		new_shape = list(field.tensor_shape) + [-1]
	else:
		new_shape = [-1]

	available_statistics = {
		'mean': np.mean,
		'max': np.max,
		'min': np.min,
		'sum': np.sum
	}

	if statistic not in available_statistics:
		raise ValueError('This statistic is not recognized.')

	if field.grid.is_regular:
		# All weights will be the same, so the array can be combined without taking the weights into account.
		return Field(available_statistics[statistic](field.reshape(tuple(reshape)), axis=tuple(axes)).reshape(tuple(new_shape)), new_grid)
	else:
		# Some weights will be different so calculate weighted mean instead.
		if statistic in ['min', 'max', 'sum']:
			f = available_statistics[statistic](field.reshape(tuple(reshape)), axis=tuple(axes))
			return Field(f.reshape(tuple(new_shape)), new_grid)
		else:
			# Statistic is mean
			weights = field.grid.weights
			w = weights.reshape(tuple(reshape)).sum(axis=tuple(axes))
			f = np.sum((field * weights).reshape(tuple(reshape)), axis=tuple(axes))
			return Field((f / w).reshape(tuple(new_shape)), new_grid)

def evaluate_supersampled(field_generator, grid, oversampling, statistic='mean', make_sparse=True):
	'''Evaluate a Field generator on `grid`, with an oversampling.

	Parameters
	----------
	field_generator : Field generator or list of Field generators
		The field generator to evaluate. If this is a list of Field generators,
		each Field generator will be evaluated and stored in a ModeBasis.
	grid : Grid
		The grid on which to evaluate `field_generator`.
	oversampling : integer or scalar or ndarray
		The factor by which to oversample. If this is a scalar, it will be rounded to
		the nearest integer. If this is an array, a different oversampling factor will
		be used for each dimension.
	statistic : string
		The statistic to compute (default is 'mean').
		The following statistics are available:
		* 'mean': compute the mean of values for points within each superpixel.
		* 'sum': compute the sum of values for points within each superpixel. This is identical to a weighted histogram.
		* 'min': compute the minimum of values for points within each superpixel.
		* 'max': compute the maximum of values for point within each superpixel.
	make_sparse : boolean
		If the resulting ModeBasis needs to be sparsified. This is ignored if
		only a single Field generator is provided.

	Returns
	-------
	Field or ModeBasis
		The evaluated field or mode basis.
	'''
	import scipy.sparse
	from ..mode_basis import ModeBasis

	if isinstance(field_generator, (list, tuple)):
		modes = []

		for fg in field_generator:
			field = evaluate_supersampled(fg, grid, oversampling, statistic)

			if make_sparse:
				field = scipy.sparse.csr_matrix(field)
				field.eliminate_zeros()

			modes.append(field)

		return ModeBasis(modes, grid)

	oversampling = (np.round(oversampling) * np.ones(grid.ndim)).astype('int')

	if grid.is_separated:
		# Use dithered sub grids to get the final supersampled field.
		deltas = []
		for i in range(grid.ndim):
			x = grid.coords.separated_coords[i]
			d = (x[2:] - x[:-2]) / 2.
			d = np.concatenate(([x[1] - x[0]], d, [x[-1] - x[-2]]))
			deltas.append(d)

		dithers = make_uniform_grid(oversampling, 1)

		separated_coords = grid.separated_coords

		if statistic in ['mean', 'sum']:
			field = 0
		else:
			field = None

		for dither in dithers.points:
			dithered_separated_coords = [c + d * delta for c, d, delta in zip(separated_coords, dither, deltas)]
			dithered_grid = grid.__class__(SeparatedCoords(dithered_separated_coords))

			if statistic in ['mean', 'sum']:
				field += field_generator(dithered_grid)
			else:
				if field is None:
					field = field_generator(dithered_grid)
				elif statistic == 'min':
					field = np.minimum(field, field_generator(dithered_grid))
				else:
					field = np.maximum(field, field_generator(dithered_grid))

		if statistic == 'mean':
			field /= len(dithers)

		field.grid = grid
		return field
	else:
		# Cannot use sub grids, so fall back to evaluation of generator on the full
		# supersampled grid.
		supersampled_grid = make_supersampled_grid(grid, oversampling)
		field = field_generator(supersampled_grid)
		return subsample_field(field, oversampling, grid, statistic)

def make_uniform_vector_field(field, jones_vector):
	'''Make an uniform vector field from a scalar field and a jones vector.

	Parameters
	----------
	field : Field
		An input scalar field that is expanded to a vector field
	jones_vector : array_like
		The output vector at every pixel

	Returns
	-------
	Field
		The expanded vector field
	'''
	if field.is_scalar_field():
		return Field([ei * field for ei in jones_vector], field.grid)

def make_uniform_vector_field_generator(field_generator, jones_vector):
	'''Make an uniform vector field generator from a scalar field generator and a jones vector.

	Parameters
	----------
	field_generator : Field generator
		The field generator to evaluate.
	jones_vector : array_like
		The output vector at every grid coordinate

	Returns
	-------
	Field generator
		This function can be evaluated on a grid to get a Field.
	'''

	def func(grid):
		scalar_field = field_generator(grid)
		return Field([ei * scalar_field for ei in jones_vector], grid)

	return func



import numpy as np
from scipy.sparse import csr_matrix
import pkg_resources

from .optical_element import OpticalElement
from ..field import make_uniform_grid, evaluate_supersampled
from ..mode_basis import ModeBasis, make_gaussian_pokes
from ..interpolation import make_linear_interpolator_separated
from ..util import read_fits

def make_actuator_positions(num_actuators_across_pupil, actuator_spacing, x_tilt=0, y_tilt=0, z_tilt=0):
	'''Make actuator positions using the BMC convention.

	Parameters
	----------
	num_actuators_across_pupil : integer
		The number of actuators across the pupil. The total number of actuators will be this number squared.
	actuator_spacing : scalar
		The spacing between actuators before tilting the deformable mirror.
	x_tilt : scalar
		The tilt of the deformable mirror around the x-axis in radians.
	y_tilt : scalar
		The tilt of the deformable mirror around the y-axis in radians.
	z_tilt : scalar
		The tilt of the deformable mirror around the z-axis in radians.

	Returns
	-------
	Grid
		The actuator positions.
	'''
	extent = actuator_spacing * num_actuators_across_pupil
	grid = make_uniform_grid(num_actuators_across_pupil, [extent, extent]).scaled(np.cos([y_tilt, x_tilt]))

	if z_tilt == 0:
		return grid

	grid = grid.rotated(z_tilt)
	return grid

def make_gaussian_influence_functions(pupil_grid, num_actuators_across_pupil, actuator_spacing, crosstalk=0.15, cutoff=3, x_tilt=0, y_tilt=0, z_tilt=0, oversampling=1):
	'''Create influence functions with a Gaussian profile.

	The default value for the crosstalk is representative for Boston Micromachines DMs.

	Parameters
	----------
	pupil_grid : Grid
		The grid on which to calculate the influence functions.
	num_actuators_across_pupil : integer
		The number of actuators across the pupil. The total number of actuators will be this number squared.
	actuator_spacing : scalar
		The spacing between actuators before tilting the deformable mirror.
	crosstalk : scalar
		The amount of crosstalk between the actuators. This is defined as the value of the influence function
		at a nearest-neighbour actuator.
	cutoff : scalar
		The distance from the center of the actuator, as a fraction of the actuator spacing, where the
		influence function is truncated to zero.
	x_tilt : scalar
		The tilt of the deformable mirror around the x-axis in radians.
	y_tilt : scalar
		The tilt of the deformable mirror around the y-axis in radians.
	z_tilt : scalar
		The tilt of the deformable mirror around the z-axis in radians.
	oversampling : integer
		The oversamping factor when creating the Gaussian. Default: 1.

	Returns
	-------
	ModeBasis
		The influence functions for each of the actuators.
	'''
	actuator_positions = make_actuator_positions(num_actuators_across_pupil, actuator_spacing)

	sigma = actuator_spacing / (np.sqrt((-2 * np.log(crosstalk))))
	cutoff = actuator_spacing / sigma * cutoff

	def transform_poke(poke):
		def new_poke(grid):
			p = poke(grid.scaled(1 / np.cos([y_tilt, x_tilt])).rotated(-z_tilt))
			p /= np.cos(x_tilt) * np.cos(y_tilt)

			return p
		return new_poke

	pokes = make_gaussian_pokes(None, actuator_positions, sigma, cutoff)
	pokes = [transform_poke(p) for p in pokes]
	pokes = evaluate_supersampled(pokes, pupil_grid, oversampling, make_sparse=True)

	return pokes

def make_xinetics_influence_functions(pupil_grid, num_actuators_across_pupil, actuator_spacing, x_tilt=0, y_tilt=0, z_tilt=0):
	'''Create influence functions for a Xinetics deformable mirror.

	This function uses a The rotation of the deformable mirror will be done in the order X-Y-Z.

	Parameters
	----------
	pupil_grid : Grid
		The grid on which to calculate the influence functions.
	num_actuators_across_pupil : integer
		The number of actuators across the pupil. The total number of actuators will be this number squared.
	actuator_spacing : scalar
		The spacing between actuators before tilting the deformable mirror.
	x_tilt : scalar
		The tilt of the deformable mirror around the x-axis in radians.
	y_tilt : scalar
		The tilt of the deformable mirror around the y-axis in radians.
	z_tilt : scalar
		The tilt of the deformable mirror around the z-axis in radians.

	Returns
	-------
	ModeBasis
		The influence functions for each of the actuators.
	'''
	actuator_positions = make_actuator_positions(num_actuators_across_pupil, actuator_spacing)

	# Stretch and rotate pupil_grid to correct for tilted DM
	evaluated_grid = pupil_grid.scaled(1 / np.cos([y_tilt, x_tilt])).rotated(-z_tilt)

	# Read in actuator shape from file.
	actuator = np.squeeze(read_fits(pkg_resources.resource_stream('hcipy', 'data/influence_dm5v2.fits')))
	actuator /= actuator.max()

	# Convert actuator into linear interpolator.
	actuator_grid = make_uniform_grid(actuator.shape, np.array(actuator.shape) * actuator_spacing / 10.0)
	actuator = make_linear_interpolator_separated(actuator.ravel(), actuator_grid, 0)

	def poke(p):
		res = csr_matrix(actuator(evaluated_grid.shifted(-p))) / np.cos(x_tilt) * np.cos(y_tilt)
		res.eliminate_zeros()

		return res

	return ModeBasis([poke(p) for p in actuator_positions.points], pupil_grid)

class DeformableMirror(OpticalElement):
	'''A deformable mirror using influence functions.

	This class does not contain any temporal simulation (ie. settling time),
	and assumes that there is no crosstalk between actuators.

	Parameters
	----------
	influence_functions : ModeBasis
		The influence function for each of the actuators.
	'''
	def __init__(self, influence_functions):
		self.influence_functions = influence_functions

		self.actuators = np.zeros(len(influence_functions))
		self._actuators_for_cached_surface = None

		self.input_grid = influence_functions.grid
		self._surface = self.input_grid.zeros()

	@property
	def num_actuators(self):
		return self._actuators.size

	@property
	def actuators(self):
		return self._actuators

	@actuators.setter
	def actuators(self, actuators):
		self._actuators = actuators

	def forward(self, wavefront):
		'''Propagate a wavefront through the deformable mirror.

		Parameters
		----------
		wavefront : Wavefront
			The incoming wavefront.

		Returns
		-------
		Wavefront
			The reflected wavefront.
		'''
		wf = wavefront.copy()
		wf.electric_field *= np.exp(2j * self.surface * wavefront.wavenumber)
		return wf

	def backward(self, wavefront):
		'''Propagate a wavefront backwards through the deformable mirror.

		Parameters
		----------
		wavefront : Wavefront
			The incoming wavefront.

		Returns
		-------
		Wavefront
			The reflected wavefront.
		'''
		wf = wavefront.copy()
		wf.electric_field *= np.exp(-2j * self.surface * wavefront.wavenumber)
		return wf

	@property
	def influence_functions(self):
		'''The influence function for each of the actuators of this deformable mirror.
		'''
		return self._influence_functions

	@influence_functions.setter
	def influence_functions(self, influence_functions):
		self._influence_functions = influence_functions
		self._actuators_for_cached_surface = None

	@property
	def surface(self):
		'''The surface of the deformable mirror in meters.
		'''
		if self._actuators_for_cached_surface is not None:
			if np.all(self.actuators == self._actuators_for_cached_surface):
				return self._surface

		self._surface = self.influence_functions.linear_combination(self.actuators)
		self._actuators_for_cached_surface = self.actuators.copy()

		return self._surface

	@property
	def opd(self):
		'''The optical path difference in meters that this deformable
		mirror induces.
		'''
		return 2 * self.surface

	def random(self, rms):
		'''Set the dm actuators with random white noise of a certain rms.

		Parameters
		----------
		rms : scalar
			The dm surface rms.
		'''
		self._actuators = np.random.randn(self._actuators.size) * rms

	def phase_for(self, wavelength):
		'''Get the phase in radians that is added to a wavefront with a specified wavelength.

		Parameters
		----------
		wavelength : scalar
			The wavelength at which to calculate the phase deformation.

		Returns
		-------
		Field
			The calculated phase deformation.
		'''
		return 2 * self.surface * 2*np.pi / wavelength

	def flatten(self):
		'''Flatten the DM by setting all actuators to zero.
		'''
		self._actuators = np.zeros(len(self.influence_functions))

def label_actuator_centroid_positions(influence_functions, label_format='{:d}', **text_kwargs):
	'''Display centroid positions for a set of influence functions.

	The location of each of the actuators is calculated using a weighted centroid, and
	at that location a label is written to the open Matplotlib Figure. The text can be
	modified with `label_format`, which is formatted with new-style Python formatting:
	`label_format.format(i)` where `i` is the actuator index.

	Parameters
	----------
	influence_functions : ModeBasis
		The influence function for each actuator.
	label_format : string
		The text that will be displayed at the actuator centroid. This must be a new-style
		formattable string.

	Raises
	------
	ValueError
		If the influence functions mode basis does not contain a grid.
	'''
	import matplotlib.pyplot as plt

	if influence_functions.grid is None:
		raise ValueError('The influence functions mode basis must contain a grid to calcuate centroids.')

	grid = influence_functions.grid.as_('cartesian')
	x, y = grid.coords

	# Center the labels unless overridden by the user.
	kwargs = {'verticalalignment': 'center', 'horizontalalignment': 'center'}
	kwargs.update(text_kwargs)

	for i, act in enumerate(influence_functions):
		x_pos = (act * x).sum() / act.sum()
		y_pos = (act * y).sum() / act.sum()
		pos = (x_pos, y_pos)

		plt.annotate(label_format.format(i), xy=pos, **kwargs)



from .wavefront_sensor import WavefrontSensorOptics, WavefrontSensorEstimator
from ..propagation import FraunhoferPropagator
from ..aperture import circular_aperture
from ..optics import SurfaceApodizer, Apodizer, TipTiltMirror
from ..field import make_pupil_grid, make_focal_grid, Field, CartesianGrid, UnstructuredCoords

import numpy as np

class ModulatedPyramidWavefrontSensorOptics(WavefrontSensorOptics):
	'''The optical elements for a modulated pyramid wavefront sensor.

	Parameters
	----------
	pyramid_wavefront_sensor : WavefrontSensorOptics
		The pyramid wavefront sensor optics that are used.
	modulation : scalar
		The modulation radius in radians.
	num_steps : int
		The number of steps per modulation cycle.
	'''
	def __init__(self, pyramid_wavefront_sensor, modulation, num_steps=12):
		self.modulation = modulation
		self.pyramid_wavefront_sensor = pyramid_wavefront_sensor
		self.tip_tilt_mirror = TipTiltMirror(self.pyramid_wavefront_sensor.input_grid)

		theta = np.linspace(0, 2 * np.pi, num_steps, endpoint=False)
		x_modulation = modulation / 2 * np.cos(theta)
		y_modulation = modulation / 2 * np.sin(theta)

		self.modulation_positions = CartesianGrid(UnstructuredCoords((x_modulation, y_modulation)))

	def forward(self, wavefront):
		'''Propagates a wavefront through the modulated pyramid wavefront sensor.

		Parameters
		----------
		wavefront : Wavefront
			The input wavefront that will propagate through the system.

		Returns
		-------
		wf_modulated : list
			A list of wavefronts for each modulation position.
		'''
		wf_modulated = []

		for point in self.modulation_positions.points:
			self.tip_tilt_mirror.actuators = point
			modulated_wavefront = self.tip_tilt_mirror.forward(wavefront)

			wf_modulated.append(self.pyramid_wavefront_sensor.forward(modulated_wavefront))

		return wf_modulated

	def backward(self, wavefront):
		raise RuntimeError('This is a non-physical operation.')

class PyramidWavefrontSensorOptics(WavefrontSensorOptics):
	'''The optical elements for a pyramid wavefront sensor.

	Parameters
	----------
	input_grid : Grid
		The grid on which the input wavefront is defined.
	separation : scalar
		The separation between the pupils. The default takes the input grid extent as separation.
	wavelength_0 : scalar
		The reference wavelength that determines the physical scales.
	q : scalar
		The focal plane oversampling coefficient. The default uses the minimal required sampling.
	refractive_index : callable
		A callable that returns the refractive index as function of wavelength.
		The default is a refractive index of 1.5.
	num_airy : scalar
		The radius of the focal plane spatial filter in units of lambda/D at the reference wavelength.
	'''
	def __init__(self, input_grid, separation=None, wavelength_0=1, q=None, num_airy=None, refractive_index=lambda x: 1.5):
		if not input_grid.is_regular:
			raise ValueError('The input grid must be a regular grid.')

		self.input_grid = input_grid
		D = np.max(input_grid.delta * (input_grid.shape - 1))

		if separation is None:
			separation = D

		# Oversampling necessary to see all frequencies in the output wavefront sensor plane
		qmin = max(2 * separation / D, 1)
		if q is None:
			q = qmin
		elif q < qmin:
			raise ValueError('The requested focal plane sampling is too low to sufficiently sample the wavefront sensor output.')

		if num_airy is None:
			self.num_airy = np.max(input_grid.shape - 1) / 2
		else:
			self.num_airy = num_airy

		self.focal_grid = make_focal_grid(q, self.num_airy, reference_wavelength=wavelength_0, pupil_diameter=D, focal_length=1)
		self.output_grid = make_pupil_grid(qmin * input_grid.dims, qmin * D)

		# Make all the optical elements
		self.spatial_filter = Apodizer(circular_aperture(2 * self.num_airy * wavelength_0 / D)(self.focal_grid))
		pyramid_surface = -separation / (2 * (refractive_index(wavelength_0) - 1)) * (np.abs(self.focal_grid.x) + np.abs(self.focal_grid.y))
		self.pyramid = SurfaceApodizer(Field(pyramid_surface, self.focal_grid), refractive_index)

		# Make the propagators
		self.pupil_to_focal = FraunhoferPropagator(input_grid, self.focal_grid)
		self.focal_to_pupil = FraunhoferPropagator(self.focal_grid, self.output_grid)

	def forward(self, wavefront):
		'''Propagates a wavefront through the pyramid wavefront sensor.

		Parameters
		----------
		wavefront : Wavefront
			The input wavefront that will propagate through the system.

		Returns
		-------
		wf_wfs : Wavefront
			The output wavefront.
		'''
		wf_focus = self.pupil_to_focal.forward(wavefront)
		wf_pyramid = self.pyramid.forward(self.spatial_filter.forward(wf_focus))
		wf_wfs = self.focal_to_pupil.forward(wf_pyramid)

		return wf_wfs

	def backward(self, wavefront):
		'''Propagates a wavefront backwards through the pyramid wavefront sensor.

		Parameters
		----------
		wavefront : Wavefront
			The input wavefront that will propagate through the system.

		Returns
		-------
		wf_pupil : Wavefront
			The output wavefront.
		'''
		wf_focus = self.focal_to_pupil.backward(wavefront)
		wf_pyramid = self.pyramid.backward(self.spatial_filter.backward(wf_focus))
		wf_pupil = self.pupil_to_focal.backward(wf_pyramid)

		return wf_pupil

class PyramidWavefrontSensorEstimator(WavefrontSensorEstimator):
	'''Estimates the wavefront slopes from pyramid wavefront sensor images.

	Parameters
	----------
	aperture : function
		A function which mask the pupils for the normalized differences.
	output_grid : Grid
		The grid on which the output of a pyramid wavefront sensor is sampled.

	Attributes
	----------
	measurement_grid : Grid
		The grid on which the normalized differences are defined.
	pupil_mask : array_like
		A mask for the normalized differences.
	num_measurements : int
		The number of pixels in the output vector.
	'''
	def __init__(self, aperture):
		self.pupil_mask = aperture
		self.num_measurements = 2 * int(np.sum(self.pupil_mask > 0))

	def estimate(self, images):
		'''A function which estimates the wavefront slope from a pyramid image.

		Parameters
		----------
		images - list
			A list of scalar intensity fields containing pyramid wavefront sensor images.

		Returns
		-------
		res - Field
			A field with wavefront sensor slopes.
		'''
		import warnings
		warnings.warn("This function does not work as expected and will be changed in a future update.", RuntimeWarning)

		image = images.shaped
		sub_shape = image.grid.shape // 2

		# Subpupils
		I_a = image[:sub_shape[0], :sub_shape[1]]
		I_b = image[sub_shape[0]:2*sub_shape[0], :sub_shape[1]]
		I_c = image[sub_shape[0]:2*sub_shape[0], sub_shape[1]:2*sub_shape[1]]
		I_d = image[:sub_shape[0], sub_shape[1]:2*sub_shape[1]]

		norm = I_a + I_b + I_c + I_d

		I_x = (I_a + I_b - I_c - I_d) / norm
		I_y = (I_a - I_b - I_c + I_d) / norm

		I_x = I_x.ravel()[self.pupil_mask>0]
		I_y = I_y.ravel()[self.pupil_mask>0]

		res = Field([I_x, I_y], self.pupil_mask.grid)
		return res



import numpy as np

def inverse_truncated_modal(M, num_modes, svd=None):
	'''Invert `M` with `num_modes` modes.

	Modes are given by a singular value decomposition of the input matrix.
	If `svd` is given, it is used instead. Modes are ordered in descending
	order by their singular values.

	Parameters
	----------
	M : ndarray
		The matrix to invert. This matrix must be two-dimensional.
	num_modes : integer
		The number of modes to use for the inversion. Higher-order modes
		are ignored.
	svd : SVD object
		A precalculated singular value decomposition object. This will
		be used if supplied, to avoid recalculation of an SVD.

	Returns
	-------
	ndarray
		The inverted matrix.
	'''
	if svd is None:
		from .singular_value_decomposition import SVD
		svd = SVD(M, num_modes)

	U, S, Vt = svd.svd

	U = U[:,:num_modes]
	S = S[:num_modes]
	Vt = Vt[:num_modes,:]

	return (Vt.T / S).dot(U.T)

def inverse_truncated(M, rcond=1e-15, svd=None):
	'''Invert `M` truncating the number of modes.

	Modes are given by a singular value decomposition of the input matrix.
	If `svd` is given, it is used instead. Modes are ordered in descending
	order by their singular values. All modes with a singular value lower than
	`rcond` times the maximum singular value, will be ignored.

	Parameters
	----------
	M : ndarray
		The matrix to invert. This matrix must be two-dimensional.
	rcond : scalar
		The relative condition number of the highest-order mode that must
		be used for inversion.
	svd : SVD object
		A precalculated singular value decomposition object. This will
		be used if supplied, to avoid recalculation of an SVD.

	Returns
	-------
	ndarray
		The inverted matrix.
	'''
	if svd is None:
		from .singular_value_decomposition import SVD
		svd = SVD(M)

	U, S, Vt = svd.svd
	S_inv = np.array([1/s if abs(s) > (rcond * S.max()) else 0 for s in S])

	return (Vt.T * S_inv).dot(U.T)

def inverse_tikhonov(M, rcond=1e-15, svd=None):
	'''Invert `M` using Tikhonov regularization.

	Tikhonov regularization will be an identity matrix, with strength
	`rcond` * the maximum singular value of the matrix.

	Parameters
	----------
	M : ndarray
		The matrix to invert. This matrix must be two-dimensional.
	rcond : scalar
		The relative strength of the regularization. An identity matrix
		will be used for the regularization.
	svd : SVD object
		A precalculated singular value decomposition object. This will
		be used if supplied, to avoid recalculation of an SVD.

	Returns
	-------
	ndarray
		The inverted matrix.
	'''
	if svd is None:
		from .singular_value_decomposition import SVD
		svd = SVD(M)

	U, S, Vt = svd.svd
	S_inv = S / (S**2 + (rcond * S.max())**2)

	return (Vt.T * S_inv).dot(U.T)



from .atmospheric_model import AtmosphericLayer, power_spectral_density_von_karman, fried_parameter_from_Cn_squared
import numpy as np
from ..util import SpectralNoiseFactoryMultiscale

class FiniteAtmosphericLayer(AtmosphericLayer):
	def __init__(self, input_grid, Cn_squared=None, L0=np.inf, velocity=0, height=0, oversampling=2):
		self._dirty = True
		self._achromatic_screen = None

		AtmosphericLayer.__init__(self, input_grid, Cn_squared, L0, velocity, height)

		self.oversampling = oversampling
		self.center = np.zeros(2)

		self.reset()

	def reset(self):
		self.psd = power_spectral_density_von_karman(fried_parameter_from_Cn_squared(self.Cn_squared, 1), self.L0)

		self.noise_factory = SpectralNoiseFactoryMultiscale(self.psd, self.input_grid, self.oversampling)
		self.noise = self.noise_factory.make_random()

		self._dirty = False

	def phase_for(self, wavelength):
		if self._dirty:
			self.reset()

		if self._achromatic_screen is None:
			self._achromatic_screen = self.noise.shifted(self.center)()

		return self._achromatic_screen / wavelength

	def evolve_until(self, t):
		self.center = self.velocity * t
		self._achromatic_screen = None

	@property
	def Cn_squared(self):
		return self._Cn_squared

	@Cn_squared.setter
	def Cn_squared(self, Cn_squared):
		self._Cn_squared = Cn_squared
		self._dirty = True

	@property
	def outer_scale(self):
		return self._L0

	@outer_scale.setter
	def L0(self, L0):
		self._L0 = L0
		self._dirty = True



from __future__ import division

import numpy as np
from .fourier_transform import FourierTransform, multiplex_for_tensor_fields, _get_float_and_complex_dtype
from ..field import Field, CartesianGrid, RegularCoords
from ..config import Configuration
import numexpr as ne

try:
	import mkl_fft._numpy_fft as _fft_module
except ImportError:
	_fft_module = np.fft

def make_fft_grid(input_grid, q=1, fov=1):
	'''Calculate the grid returned by a Fast Fourier Transform.

	Parameters
	----------
	input_grid : Grid
		The grid defining the sampling in the real domain..
	q : scalar
		The amount of zeropadding to perform. A value of 1 denotes no zeropadding.
	fov : scalar
		The amount of cropping to perform in the Fourier domain.

	Returns
	-------
	Grid
		The grid defining the sampling in the Fourier domain.
	'''
	q = np.ones(input_grid.ndim, dtype='float') * q
	fov = np.ones(input_grid.ndim, dtype='float') * fov

	# Check assumptions
	if not input_grid.is_regular:
		raise ValueError('The input_grid must be regular.')
	if not input_grid.is_('cartesian'):
		raise ValueError('The input_grid must be cartesian.')

	delta = (2 * np.pi / (input_grid.delta * input_grid.dims)) / q
	dims = (input_grid.dims * fov * q).astype('int')
	zero = delta * (-dims / 2 + np.mod(dims, 2) * 0.5)

	return CartesianGrid(RegularCoords(delta, dims, zero))

def _numexpr_grid_shift(shift, grid, out=None):
	'''Fast evaluation of np.exp(1j * np.dot(shift, grid.coords)) using NumExpr.

	Parameters
	----------
	shift : array_like
		The coordinates of the shift.
	grid : Grid
		The grid on which to calculate the shift.
	out : array_like
		An existing array where the outcome is going to be stored. This must
		have the correct shape and dtype. No checking will be performed. If
		this is None, a new array will be allocated and returned.

	Returns
	-------
	array_like
		The calculated complex shift array.
	'''
	variables = {}
	command = []
	coords = grid.coords

	for i in range(grid.ndim):
		variables['a%d' % i] = shift[i]
		variables['b%d' % i] = coords[i]

		command.append('a%d * b%d' % (i,i))

	command = 'exp(1j * (' + '+'.join(command) + '))'
	return ne.evaluate(command, local_dict=variables, out=out)

class FastFourierTransform(FourierTransform):
	'''A Fast Fourier Transform (FFT) object.

	This Fourier transform calculates FFTs with zeropadding and cropping. This
	Fourier transform requires the input grid to be regular in Cartesian coordinates. Every
	number of dimensions is allowed.

	Parameters
	----------
	input_grid : Grid
		The grid that is expected for the input field.
	q : scalar
		The amount of zeropadding to perform. A value of 1 denotes no zeropadding.
	fov : scalar
		The amount of cropping to perform in the Fourier domain.
	shift : array_like or scalar
		The amount by which to shift the output grid.
	emulate_fftshifts : boolean or None
		Whether to emulate FFTshifts normally used in the FFT by multiplications in the
		opposite domain. Enabling this increases performance by 3x, but degrades accuracy of
		the FFT by 10x. If this is None, the choice will be determined by the configuration
		file.

	Raises
	------
	ValueError
		If the input grid is not regular or Cartesian.
	'''
	def __init__(self, input_grid, q=1, fov=1, shift=0, emulate_fftshifts=None):
		# Check assumptions
		if not input_grid.is_regular:
			raise ValueError('The input_grid must be regular.')
		if not input_grid.is_('cartesian'):
			raise ValueError('The input_grid must be Cartesian.')

		self.input_grid = input_grid

		self.shape_in = input_grid.shape
		self.weights = input_grid.weights
		self.size = input_grid.size
		self.ndim = input_grid.ndim

		# Get the value from the configuration file if left at the default.
		if emulate_fftshifts is None:
			emulate_fftshifts = Configuration().fourier.fft.emulate_fftshifts
		self.emulate_fftshifts = emulate_fftshifts

		self.output_grid = make_fft_grid(input_grid, q, fov).shifted(shift)
		self.internal_grid = make_fft_grid(input_grid, q, 1)

		self.shape_out = self.output_grid.shape
		self.internal_shape = (self.shape_in * q).astype('int')
		self.internal_array = np.zeros(self.internal_shape, 'complex')

		# Calculate the part of the array in which to insert the input field (for zeropadding).
		if np.allclose(self.internal_shape, self.shape_in):
			self.cutout_input = None
		else:
			cutout_start = (self.internal_shape / 2.).astype('int') - (self.shape_in / 2.).astype('int')
			cutout_end = cutout_start + self.shape_in
			self.cutout_input = tuple([slice(start, end) for start, end in zip(cutout_start, cutout_end)])

		# Calculate the part of the array to extract the output field (for cropping).
		if np.allclose(self.internal_shape, self.shape_out):
			self.cutout_output = None
		else:
			cutout_start = (self.internal_shape / 2.).astype('int') - (self.shape_out / 2.).astype('int')
			cutout_end = cutout_start + self.shape_out
			self.cutout_output = tuple([slice(start, end) for start, end in zip(cutout_start, cutout_end)])

		# Calculate the shift array when the input grid was shifted compared to the native shift
		# expected by the numpy FFT implementation.
		center = input_grid.zero + input_grid.delta * (np.array(input_grid.dims) // 2)
		self.shift_input = _numexpr_grid_shift(-center, self.output_grid)

		# Remove piston shift (remove central shift phase)
		self.shift_input /= np.fft.ifftshift(self.shift_input.reshape(self.shape_out)).ravel()[0]

		# Calculate the multiplication for emulating the FFTshift (if requested).
		if emulate_fftshifts:
			f_shift = input_grid.delta * np.array(self.internal_shape[::-1] // 2)
			fftshift = _numexpr_grid_shift(f_shift, self.internal_grid)

			if self.cutout_output:
				self.shift_input *= fftshift.reshape(self.internal_shape)[self.cutout_output].ravel()
			else:
				self.shift_input *= fftshift

		# Apply weights for Fourier normalization.
		self.shift_input *= self.weights

		# Calculate the shift array when the output grid was shifted compared to the native shift
		# expcted by the numpy FFT implementation.
		shift = np.ones(self.input_grid.ndim) * shift
		if np.allclose(shift, 0):
			self.shift_output = 1
		else:
			self.shift_output = _numexpr_grid_shift(-shift, self.input_grid)

		# Calculate the multiplication for emulating the FFTshift (if requested).
		if emulate_fftshifts:
			f_shift = self.input_grid.delta * np.array(self.internal_shape[::-1] // 2)
			fftshift = _numexpr_grid_shift(f_shift, self.internal_grid)

			fftshift *= np.exp(-1j * np.dot(f_shift, self.internal_grid.zero))

			if self.cutout_input:
				self.shift_output *= fftshift.reshape(self.internal_shape)[self.cutout_input].ravel()
			else:
				self.shift_output *= fftshift

		# Detect if we don't need to shift in the output plane (to avoid a multiplication in the operation).
		if np.isscalar(self.shift_output) and np.allclose(self.shift_output, 1):
			self.shift_output = None

	@multiplex_for_tensor_fields
	def forward(self, field):
		'''Returns the forward Fourier transform of the :class:`Field` field.

		Parameters
		----------
		field : Field
			The field to Fourier transform.

		Returns
		-------
		Field
			The Fourier transform of the field.
		'''
		if self.cutout_input is None:
			self.internal_array[:] = field.reshape(self.shape_in)

			if self.shift_output is not None:
				self.internal_array *= self.shift_output.reshape(self.shape_in)
		else:
			self.internal_array[:] = 0
			self.internal_array[self.cutout_input] = field.reshape(self.shape_in)

			if self.shift_output is not None:
				self.internal_array[self.cutout_input] *= self.shift_output.reshape(self.shape_in)

		if not self.emulate_fftshifts:
			self.internal_array = np.fft.ifftshift(self.internal_array)

		fft_array = _fft_module.fftn(self.internal_array)

		if not self.emulate_fftshifts:
			fft_array = np.fft.fftshift(fft_array)

		if self.cutout_output is None:
			res = fft_array.ravel()
		else:
			res = fft_array[self.cutout_output].ravel()

		res *= self.shift_input

		float_dtype, complex_dtype = _get_float_and_complex_dtype(field.dtype)
		return Field(res, self.output_grid).astype(complex_dtype, copy=False)

	@multiplex_for_tensor_fields
	def backward(self, field):
		'''Returns the inverse Fourier transform of the :class:`Field` field.

		Parameters
		----------
		field : Field
			The field to inverse Fourier transform.

		Returns
		-------
		Field
			The inverse Fourier transform of the field.
		'''
		if self.cutout_output is None:
			self.internal_array[:] = field.reshape(self.shape_out)
			self.internal_array /= self.shift_input.reshape(self.shape_out)
		else:
			self.internal_array[:] = 0
			self.internal_array[self.cutout_output] = field.reshape(self.shape_out)
			self.internal_array[self.cutout_output] /= self.shift_input.reshape(self.shape_out)

		if not self.emulate_fftshifts:
			self.internal_array = np.fft.ifftshift(self.internal_array)

		fft_array = _fft_module.ifftn(self.internal_array)

		if not self.emulate_fftshifts:
			fft_array = np.fft.fftshift(fft_array)

		if self.cutout_input is None:
			res = fft_array.ravel()
		else:
			res = fft_array[self.cutout_input].ravel()

		if self.shift_output is not None:
				res /= self.shift_output

		float_dtype, complex_dtype = _get_float_and_complex_dtype(field.dtype)
		return Field(res, self.input_grid).astype(complex_dtype, copy=False)



import numpy as np

from ..field import Field
from ..util import SpectralNoiseFactoryFFT, inverse_tikhonov
from .apodization import SurfaceApodizer
from ..propagation import FresnelPropagator
from ..aperture import circular_aperture
from .optical_element import OpticalElement

def make_power_law_error(pupil_grid, ptv, diameter, exponent=-2.5, aperture=None, remove_modes=None):
	'''Create an error surface from a power-law power spectral density.

	Parameters
	----------
	pupil_grid : Grid
		The grid on which to calculate the error.
	ptv : scalar
		The peak-to-valley of the wavefront aberration in meters.
	diameter : scalar
		The diameter over which the ptv is calculated.
	exponent : scalar
		The exponent of the power law.
	aperture : Field
		The mask over which to calculate the ptv. A circular aperture with diameter
		`diameter` is used if this is not given.
	remove_modes : ModeBasis
		The modes which to remove from the surface aberration. The peak-to-valley
		is enforced before these modes are removed. This allows for correctting surface
		errors with optic alignment.

	Returns
	-------
	Field
		The surface error calculated on `pupil_grid`.
	'''
	def psd(grid):
		res = Field(grid.as_('polar').r**exponent, grid)
		res[grid.as_('polar').r == 0] = 0
		return res

	if aperture is None:
		aperture = circular_aperture(diameter)(pupil_grid)

	screen = SpectralNoiseFactoryFFT(psd, pupil_grid).make_random()()
	screen *= ptv / np.ptp(screen[aperture != 0])

	if remove_modes is not None:
		trans = remove_modes.transformation_matrix
		trans_inv = inverse_tikhonov(trans, 1e-6)
		screen -= trans.dot(trans_inv.dot(screen))

	return Field(screen * aperture, pupil_grid)

class SurfaceAberration(SurfaceApodizer):
	'''A surface aberration with a specific power law.

	Parameters
	----------
	pupil_grid : Grid
		The grid on which the incoming wavefront is defined.
	ptv : scalar
		The peak-to-valley of the wavefront aberration in meters.
	diameter : scalar
		The diameter over which the ptv is calculated.
	exponent : scalar
		The exponent of the power law.
	refractive_index : scalar
		The refractive index of the surface for which this is the surface error.
		The default is a mirror surface.
	aperture : Field
		The mask over which to calculate the ptv. A circular aperture with diameter
		`diameter` is used if this is not given.
	remove_modes : ModeBasis
		The modes which to remove from the surface aberration. The peak-to-valley
		is enforced before these modes are removed. This allows for correctting surface
		errors with optic alignment.
	'''
	def __init__(self, pupil_grid, ptv, diameter, exponent=-2.5, refractive_index=-1, aperture=None, remove_modes=None):
		surface = make_power_law_error(pupil_grid, ptv, diameter, exponent, aperture, remove_modes)
		SurfaceApodizer.__init__(self, surface, refractive_index)

class SurfaceAberrationAtDistance(OpticalElement):
	'''A surface at a certain distance from the current plane.

	Light is propagated to this surface, then the surface errors are
	applied, and afterwards the light is propagated back towards the
	original plane. This allows for easy addition of surface errors on
	lenses, while still retaining the Fraunhofer propagations in between
	focal and pupil planes.

	Parameters
	----------
	surface_aberration : OpticalElement
		The optical element describing the surface aberration.
	distance : scalar
		The distance from the current plane.
	'''
	def __init__(self, surface_aberration, distance):
		self.fresnel = FresnelPropagator(surface_aberration.input_grid, distance)
		self.surface_aberration = surface_aberration

	def forward(self, wavefront):
		'''Propagate a wavefront forwards through the surface aberration.

		Parameters
		----------
		wavefront : Wavefront
			The incoming wavefront.

		Returns
		-------
		Wavefront
			The wavefront after the surface aberration. This wavefront is
			given at the same plane as `wavefront`.
		'''
		wf = self.fresnel.forward(wavefront)
		wf = self.surface_aberration.forward(wf)
		return self.fresnel.backward(wf)

	def backward(self, wavefront):
		'''Propagate a wavefront backwards through the surface aberration.

		Parameters
		----------
		wavefront : Wavefront
			The incoming wavefront.

		Returns
		-------
		Wavefront
			The wavefront before the surface aberration. This wavefront is
			given at the same plane as `wavefront`.
		'''
		wf = self.fresnel.forward(wavefront)
		wf = self.surface_aberration.backward(wf)
		return self.fresnel.backward(wf)



from hcipy import *
import numpy as np
import matplotlib.pyplot as plt

def test_optical_differentiation_wavefront_sensor():
	pupil_grid = make_pupil_grid(128, 1)
	wfs_grid = make_pupil_grid(256, 2)
	amplitude_filter = make_polarization_odwfs_amplitude_filter(0.0)

	odwfs = OpticalDifferentiationWavefrontSensorOptics(amplitude_filter, pupil_grid, wfs_grid)

	zernike_modes = make_zernike_basis(20, 1, pupil_grid)
	aberration = zernike_modes.linear_combination(np.random.randn(20)) * 0.1

	wf = Wavefront(circular_aperture(1)(pupil_grid) * np.exp(1j * aberration))
	img = odwfs(wf).intensity

def test_pyramid_wavefront_sensor():
	pupil_grid = make_pupil_grid(128)

	pywfs = PyramidWavefrontSensorOptics(pupil_grid, 1.2)

	zernike_modes = make_zernike_basis(20, 1, pupil_grid)
	aberration = zernike_modes.linear_combination(np.random.randn(20)) * 0.1

	wf = Wavefront(circular_aperture(1)(pupil_grid) * np.exp(1j * aberration))
	img = pywfs(wf).intensity

def test_modulated_pyramid_wavefront_sensor():
	pupil_grid = make_pupil_grid(128)

	pywfs = PyramidWavefrontSensorOptics(pupil_grid, 1.2)
	mpywfs = ModulatedPyramidWavefrontSensorOptics(pywfs, 2)

	zernike_modes = make_zernike_basis(20, 1, pupil_grid)
	aberration = zernike_modes.linear_combination(np.random.randn(20)) * 0.1

	wf = Wavefront(circular_aperture(1)(pupil_grid) * np.exp(1j * aberration))
	imgs = mpywfs(wf)
	img = np.sum([im.intensity for im in mpywfs(wf)], axis=0)
	img = Field(img, imgs[0].grid)

def test_zernike_wavefront_sensor():
	pupil_grid = make_pupil_grid(128)

	zwfs = ZernikeWavefrontSensorOptics(pupil_grid)

	zernike_modes = make_zernike_basis(20, 1, pupil_grid)
	aberration = zernike_modes.linear_combination(np.random.randn(20)) * 0.1

	wf = Wavefront(circular_aperture(1)(pupil_grid) * np.exp(1j * aberration))
	img = zwfs(wf).intensity

def test_vector_zernike_wavefront_sensor():
	pupil_grid = make_pupil_grid(128)

	vzwfs = VectorZernikeWavefrontSensorOptics(pupil_grid)

	zwfs_plus = ZernikeWavefrontSensorOptics(pupil_grid, np.pi / 2)
	zwfs_neg = ZernikeWavefrontSensorOptics(pupil_grid, -np.pi / 2)

	zernike_modes = make_zernike_basis(20, 1, pupil_grid)
	aberration = zernike_modes.linear_combination(np.random.randn(20)) * 0.1

	wf = Wavefront(circular_aperture(1)(pupil_grid) * np.exp(1j * aberration))

	vector_img = vzwfs(wf).intensity
	ref_img = (zwfs_plus(wf).intensity + zwfs_neg(wf).intensity) / 2

	assert np.allclose(vector_img, ref_img)



from .wavefront_sensor import WavefrontSensorOptics, WavefrontSensorEstimator
from ..propagation import FraunhoferPropagator
from ..optics import MultiplexedComplexSurfaceApodizer
from ..field import make_pupil_grid, Field

import numpy as np

def make_polarization_odwfs_amplitude_filter(beta):
	'''
	The OD-wfs amplitude filter based on polarization optics following [Haffert2016]_.

	.. [Haffert2016] S. Y. Haffert, 2016, "Generalised optical differentiation wavefront sensor:
		a sensitive high dynamic range wavefront sensor," Opt. Express 24, 18986-19007 (2016).

	Parameters
	----------
	beta : scalar

	Returns
	-------
	function
		A function that generates the amplitude profile.

	Raises
	------
	ValueError
		If beta is not between 0 and 1.
	'''
	if beta < 0 or beta > 1:
		raise ValueError('Beta should be between 0 and 1.')

	return lambda x: np.sin(np.pi / 2 * (beta * np.heaviside(x, 0.5) + (1 - beta) * (1 + x) / 2))

def make_odwfs_amplitude_filter(beta):
	'''
	The OD-wfs amplitude filter following [Haffert2016]_.

	.. [Haffert2016] S. Y. Haffert, 2016, "Generalised optical differentiation wavefront sensor:
		a sensitive high dynamic range wavefront sensor," Opt. Express 24, 18986-19007 (2016).

	Parameters
	----------
	beta : scalar

	Returns
	-------
	function
		A function that generates the amplitude profile.

	Raises
	------
	ValueError
		If beta is not between 0 and 1.
	'''
	if beta < 0 or beta > 1:
		raise ValueError('Beta should be between 0 and 1.')

	return lambda x: (beta * np.heaviside(x, 0.5) + (1 - beta) * (1 + x) / 2) / np.sqrt(2)

def optical_differentiation_surface(filter_size, amplitude_filter, separation, wavelength_0, refractive_index, orientation=0):
	'''A generator function for the complex multiplexed surface of the ODWFS.

	Parameters
	----------
	filter_size : scalar
		The physical size of the filter in lambda/D.
	amplitude_filter : lambda function
		A lambda function which acts on the focal plane grid to create a amplitude filter.
	separation : scalar
		The separation of the pupils in pupil diameters.
	wavelength_0 : scalar
		The reference wavelength for the filter specifications.
	refractive_index : lambda function
		A lambda function for the refractive index which accepts a wavelength.
	orientation : scalar
		The orientation of the amplitude filters. The default value is 0 radian.
	Returns
	----------
	function
		The returned function acts on a grid to create the amplitude filters for that grid.
	'''
	def func(grid):
		surface_grid = grid.rotated(orientation)

		surface_tilt = 0.5 * separation / (refractive_index(wavelength_0) - 1)

		# The surfaces which tilt the beam
		surf1 = -surface_tilt * surface_grid.x
		surf2 = -surface_tilt * -surface_grid.x
		surf3 = -surface_tilt * surface_grid.y
		surf4 = -surface_tilt * -surface_grid.y

		surf = (Field(surf1, surface_grid), Field(surf2, surface_grid), Field(surf3, surface_grid), Field(surf4, surface_grid))

		# The physical boundaries of the mask
		filter_mask = (np.abs(grid.x) < filter_size) * (np.abs(grid.y) < filter_size)

		# Note: be careful with the plus and minus signs of the filters.
		# For energy conservation the squared sum of the filters should be <= 1
		# For electric field conservation the second filter has to have opposite sign.
		filter_1 = amplitude_filter(surface_grid.x / filter_size)
		filter_1 *= filter_mask

		filter_2 = -amplitude_filter(-surface_grid.x / filter_size)
		filter_2 *= filter_mask

		filter_3 = amplitude_filter(surface_grid.y / filter_size)
		filter_3 *= filter_mask
		for f in filter_3:
			print(f)

		filter_4 = -amplitude_filter(-surface_grid.y / filter_size)
		filter_4 *= filter_mask

		amp = (Field(filter_1, grid), Field(filter_2, grid), Field(filter_3, grid), Field(filter_4, grid))

		return MultiplexedComplexSurfaceApodizer(amp, surf, refractive_index)
	return func

class OpticalDifferentiationWavefrontSensorOptics(WavefrontSensorOptics):
	'''The optical elements for a optical-differentiation wavefront sensor.

	Parameters
	----------
	amplitude_filter : callable
		The function that defines the amplitude filter in the focal plane.
	input_grid : Grid
		The grid on which the input wavefront is defined.
	output_grid : Grid
		The grid on which the output wavefront is defined.
	separation : scalar
		The separation between the pupils. The default takes the input grid extent as separation.
	D : scalar
		The size of the pupil. The default take sthe input grid extent as pupil size.
	wavelength_0 : scalar
		The reference wavelength that determines the physical scales.
	q : scalar
		The focal plane oversampling coefficient. The default uses the minimal required sampling.
	num_airy : scalar
		The radius of the focal plane spatial filter in units of lambda/D at the reference wavelength.
	refractive_index : callable
		A callable that returns the refractive index as function of wavelength.
		The default is a refractive index of 1.5.
	'''
	def __init__(self, amplitude_filter, input_grid, output_grid, separation=None, D=None, wavelength_0=1, q=None, num_airy=None, refractive_index=lambda x: 1.5):
		if not input_grid.is_regular:
			raise ValueError('The input grid must be a regular grid.')

		self.input_grid = input_grid
		self.output_grid = output_grid

		if D is None:
			D = np.max(input_grid.delta * (input_grid.shape - 1))

		if separation is None:
			separation = D

		# Create the intermediate focal grid
		# Oversampling necessary to see all frequencies in the output wavefront sensor plane
		qmin = np.ceil(max(output_grid.x.ptp() / input_grid.x.ptp(), 2))
		if q is None:
			q = qmin
		elif q < qmin:
			raise ValueError('The requested focal plane sampling is too low to sufficiently sample the wavefront sensor output.')

		if num_airy is None:
			self.num_airy = np.max(input_grid.shape - 1) / 2
		else:
			self.num_airy = num_airy

		num_pixels = 2 * int(self.num_airy * q)
		spatial_resolution = wavelength_0 / D
		self.focal_grid = make_pupil_grid(num_pixels, 2 * spatial_resolution * self.num_airy)

		# Make all the optical elements
		self.filter_size = self.num_airy * wavelength_0 / D
		focal_plane_mask = optical_differentiation_surface(self.filter_size, amplitude_filter, separation * np.sqrt(2), wavelength_0, refractive_index, orientation=np.pi/4)
		self.focal_mask = focal_plane_mask(self.focal_grid)

		# Make the propagators
		self.pupil_to_focal = FraunhoferPropagator(self.input_grid, self.focal_grid)
		self.focal_to_pupil = FraunhoferPropagator(self.focal_grid, self.output_grid)

	def forward(self, wavefront):
		'''Propagates a wavefront through the wavefront sensor.

		Parameters
		----------
		wavefront : Wavefront
			The input wavefront that will propagate through the system.

		Returns
		-------
		Wavefront
			The output wavefront.
		'''
		wf = self.pupil_to_focal.forward(wavefront)
		wf = self.focal_mask.forward(wf)
		wf = self.focal_to_pupil.forward(wf)

		return wf

	def backward(self, wavefront):
		'''Propagates a wavefront backwards through the wavefront sensor.

		Parameters
		----------
		wavefront : Wavefront
			The input wavefront that will propagate through the system.

		Returns
		-------
		Wavefront
			The output wavefront.
		'''
		wf = self.pupil_to_focal.backward(wavefront)
		wf = self.focal_mask.backward(wf)
		wf = self.focal_to_pupil.backward(wf)

		return wf

class OpticalDifferentiationWavefrontSensorEstimator(WavefrontSensorEstimator):
	'''Estimates the wavefront slopes from pyramid wavefront sensor images.

	Parameters
	----------
	aperture : function
		A function which mask the pupils for the normalized differences.
	output_grid : Grid
		The grid on which the output of a pyramid wavefront sensor is sampled.

	Attributes
	----------
	measurement_grid : Grid
		The grid on which the normalized differences are defined.
	pupil_mask : array_like
		A mask for the normalized differences.
	num_measurements : int
		The number of pixels in the output vector.
	'''
	def __init__(self, slope_grid, aperture):
		self.pupil_mask = aperture
		self.num_measurements = 2 * int(np.sum(self.pupil_mask > 0))

	def estimate(self, images):
		'''A function which estimates the wavefront slope from a pyramid image.

		Parameters
		----------
		images : list
			A list of scalar intensity fields containing pyramid wavefront sensor images.

		Returns
		-------
		array_like
			The wavefront sensor slopes.
		'''
		import warnings
		warnings.warn("This function does not work as expected and will be changed in a future update.", RuntimeWarning)

		image = images.shaped
		sub_shape = np.array(image.grid.shape) // 2

		# Subpupils
		I_a = image[:sub_shape[0], :sub_shape[1]]
		I_b = image[sub_shape[0]:2*sub_shape[0], :sub_shape[1]]
		I_c = image[sub_shape[0]:2*sub_shape[0], sub_shape[1]:2*sub_shape[1]]
		I_d = image[:sub_shape[0], sub_shape[1]:2*sub_shape[1]]

		I_x = (I_a - I_c) / (I_a + I_c)
		I_y = (I_b - I_d) / (I_b + I_d)

		I_x = I_x.ravel()[self.pupil_mask>0]
		I_y = I_y.ravel()[self.pupil_mask>0]

		slopes = np.vstack((I_x, I_y))
		return slopes


import numpy as np
import copy

from ..field import Field

class SpectralNoiseFactory(object):
	def __init__(self, psd, output_grid, psd_args=(), psd_kwargs=None):
		if psd_kwargs is None:
			psd_kwargs = {}

		self.psd = psd
		self.psd_args = psd_args
		self.psd_kwargs = psd_kwargs
		self.output_grid = output_grid

	def make_random(self):
		raise NotImplementedError()

	def make_zero(self):
		raise NotImplementedError()

class SpectralNoise(object):
	def copy(self):
		return copy.deepcopy(self)

	def shift(self, shift):
		raise NotImplementedError()

	def shifted(self, shift):
		a = self.copy()
		a.shift(shift)
		return a

	def __iadd__(self, b):
		return NotImplemented

	def __add__(self, b):
		a = self.copy()
		a += b
		return a

	def __imul__(self, f):
		return NotImplemented

	def __mul__(self, f):
		a = self.copy()
		a *= f
		return a

	def __call__(self):
		raise NotImplementedError()

	def evaluate(self):
		return self()

class SpectralNoiseFactoryFFT(SpectralNoiseFactory):
	def __init__(self, psd, output_grid, oversample=1, psd_args=(), psd_kwargs=None):
		from ..fourier import FastFourierTransform

		if psd_kwargs is None:
			psd_kwargs = {}

		SpectralNoiseFactory.__init__(self, psd, output_grid, psd_args, psd_kwargs)

		if not self.output_grid.is_regular:
			raise ValueError("Can't make a SpectralNoiseFactoryFFT on a non-regular grid.")

		self.fourier = FastFourierTransform(self.output_grid, q=oversample)
		self.input_grid = self.fourier.output_grid

		self.period = output_grid.coords.delta * output_grid.coords.shape

		# * (2*np.pi)**self.input_grid.ndim is due to conversion from PSD from "per Hertz" to "per radian", which yields a factor of 2pi per dimension
		self.C = np.sqrt(self.psd(self.input_grid, *psd_args, **psd_kwargs) / self.input_grid.weights * (2*np.pi)**self.input_grid.ndim)

	def make_random(self):
		N = self.input_grid.size
		C = self.C * (np.random.randn(N) + 1j * np.random.randn(N))
		C = Field(C, self.input_grid)

		return SpectralNoiseFFT(self, C)

	def make_zero(self):
		C = Field(np.zeros(self.input_grid.size, dtype='complex'), self.input_grid)
		return SpectralNoiseFFT(self, C)

class SpectralNoiseFFT(SpectralNoise):
	def __init__(self, factory, C):
		self.factory = factory
		self.C = C
		self.coords = C.grid.separated_coords

	def shift(self, shift):
		s = shift#np.mod(shift + self.factory.period / 2, self.factory.period) - self.factory.period / 2
		S = [s[i] * self.coords[i] for i in range(len(self.coords))]
		S = np.add.reduce(np.ix_(*S))
		self.C *= np.exp(-1j * S.ravel())

	def __iadd__(self, b):
		self.C += b.C
		return self

	def __imul__(self, f):
		self.C *= f
		return self

	def __call__(self):
		return self.factory.fourier.backward(self.C).real

class SpectralNoiseFactoryMultiscale(SpectralNoiseFactory):
	def __init__(self, psd, output_grid, oversampling, psd_args=(), psd_kwargs=None):
		from ..fourier import FastFourierTransform, MatrixFourierTransform

		if psd_kwargs is None:
			psd_kwargs = {}

		SpectralNoiseFactory.__init__(self, psd, output_grid, psd_args, psd_kwargs)

		self.oversampling = oversampling

		self.fourier_1 = FastFourierTransform(self.output_grid)
		self.input_grid_1 = self.fourier_1.output_grid

		self.input_grid_2 = self.input_grid_1.scaled(1.0 / oversampling)
		self.fourier_2 = MatrixFourierTransform(self.output_grid, self.input_grid_2)

		boundary = np.abs(self.input_grid_2.x).max()
		mask_1 = self.input_grid_1.as_('polar').r < boundary
		mask_2 = self.input_grid_2.as_('polar').r >= boundary

		# * (2*np.pi)**self.input_grid.ndim is due to conversion from PSD from "per Hertz" to "per radian", which yields a factor of 2pi per dimension
		self.C_1 = np.sqrt(psd(self.input_grid_1, *psd_args, **psd_kwargs) / self.input_grid_1.weights * (2*np.pi)**self.input_grid_1.ndim)
		self.C_1[mask_1] = 0
		self.C_2 = np.sqrt(psd(self.input_grid_2, *psd_args, **psd_kwargs) / self.input_grid_2.weights * (2*np.pi)**self.input_grid_1.ndim)
		self.C_2[mask_2] = 0

	def make_random(self):
		N_1 = self.input_grid_1.size
		N_2 = self.input_grid_2.size

		C_1 = self.C_1 * (np.random.randn(N_1) + 1j * np.random.randn(N_1))
		C_2 = self.C_2 * (np.random.randn(N_2) + 1j * np.random.randn(N_2))

		return SpectralNoiseMultiscale(self, C_1, C_2)

	def make_zero(self):
		N_1 = self.input_grid_1.size
		N_2 = self.input_grid_2.size

		C_1 = Field(np.zeros(N_1, dtype='complex'), self.C_1.grid)
		C_2 = Field(np.zeros(N_2, dtype='complex'), self.C_2.grid)

		return SpectralNoiseMultiscale(self, C_1, C_2)

class SpectralNoiseMultiscale(SpectralNoise):
	def __init__(self, factory, C_1, C_2):
		self.factory = factory

		self.C_1 = C_1
		self.C_2 = C_2

		self.coords_1 = C_1.grid.separated_coords
		self.coords_2 = C_2.grid.separated_coords

	def shift(self, shift):
		S_1 = [shift[i] * self.coords_1[i] for i in range(len(self.coords_1))]
		S_1 = sum(np.ix_(*S_1))

		S_2 = [shift[i] * self.coords_2[i] for i in range(len(self.coords_2))]
		S_2 = sum(np.ix_(*S_2))

		self.C_1 *= np.exp(-1j * S_1.ravel())
		self.C_2 *= np.exp(-1j * S_2.ravel())

	def __iadd__(self, b):
		self.C_1 += b.C_1
		self.C_2 += b.C_2
		return self

	def __imul__(self, f):
		self.C_1 *= f
		self.C_2 *= f
		return self

	def __call__(self):
		ps = self.factory.fourier_1.backward(self.C_1).real
		ps += self.factory.fourier_2.backward(self.C_2).real
		return ps



import matplotlib
matplotlib.use('Agg')

from hcipy import *
import numpy as np
import matplotlib.pyplot as plt
import os
import pytest

def test_gif_writer():
	grid = make_pupil_grid(256)

	mw = GifWriter('test.gif')

	for i in range(25):
		field = Field(np.random.randn(grid.size), grid)

		plt.clf()
		imshow_field(field)

		mw.add_frame()

	mw.close()

	assert os.path.isfile('test.gif')
	assert not os.path.exists('test.gif.frames')

	pytest.raises(RuntimeError, mw.add_frame)

	os.remove('test.gif')

def test_imshow_field():
	grid = make_pupil_grid(256)

	field = Field(np.random.randn(grid.size), grid)

	imshow_field(field)
	plt.clf()

	mask = circular_aperture(1)(grid)

	imshow_field(field, mask=mask)
	plt.clf()

	field = Field(np.random.randn(grid.size) + 1j * np.random.randn(grid.size), grid)

	imshow_field(field)
	plt.clf()

def test_imsave_field():
	grid = make_pupil_grid(256)

	field = Field(np.random.randn(grid.size), grid)

	imsave_field('field.png', field)
	assert os.path.isfile('field.png')

	os.remove('field.png')

def test_contour_field():
	grid = make_pupil_grid(256)

	field = Field(np.random.randn(grid.size), grid)

	contour_field(field)
	plt.clf()

	contourf_field(field)
	plt.clf()

def test_imshow_util():
	pupil_grid = make_pupil_grid(128)
	focal_grid = make_focal_grid(4, 16)

	aperture = make_magellan_aperture(True)(pupil_grid)
	prop = FraunhoferPropagator(pupil_grid, focal_grid)

	wf = Wavefront(aperture)
	wf.electric_field *= np.exp(0.1j * zernike(6, 2, radial_cutoff=False)(pupil_grid))

	imshow_pupil_phase(wf, remove_piston=True, crosshairs=True, title='phase')
	plt.clf()

	img = prop(wf)

	imshow_psf(img, colorbar_orientation='vertical', normalization='peak', crosshairs=True, title='psf')
	plt.clf()

	imshow_psf(img, scale='linear', colorbar_orientation='vertical', normalization='peak', crosshairs=True, title='psf')
	plt.clf()



import numpy as np
from scipy.signal import windows

from ..optics import OpticalElement, LinearRetarder, Apodizer, AgnosticOpticalElement, make_agnostic_forward, make_agnostic_backward, Wavefront
from ..propagation import FraunhoferPropagator
from ..field import make_focal_grid, Field, field_dot
from ..aperture import circular_aperture
from ..fourier import FastFourierTransform, MatrixFourierTransform, FourierFilter

class VortexCoronagraph(OpticalElement):
	'''An optical vortex coronagraph.

	This :class:`OpticalElement` simulations the propagation of light through
	a vortex in the focal plane. To resolve the singularity of this vortex
	phase plate, a multi-scale approach is made. Discretisation errors made at
	a certain level are corrected by the next level with finer sampling.

	Parameters
	----------
	input_grid : Grid
		The grid on which the incoming wavefront is defined.
	charge : integer
		The charge of the vortex.
	lyot_stop : Field or OpticalElement
		The Lyot stop for the coronagraph. If it's a Field, it is converted to an
		OpticalElement for convenience. If this is None (default), then no Lyot stop is used.
	q : scalar
		The minimum number of pixels per lambda/D. The number of levels in the multi-scale
		Fourier transforms will be chosen to reach at least this number of samples. The required
		q for a high-accuracy vortex coronagraph depends on the charge of the vortex. For charge 2,
		this can be as low as 32, but for charge 8 you need ~1024. Lower values give higher performance
		as a smaller number of levels is needed, but increases the sampling errors near the singularity.
		Charges not divisible by four require a much lower q. The default (q=1024) is conservative in
		most cases.
	scaling_factor : scalar
		The fractional increase in spatial frequency sampling per level. Larger scaling factors
		require a smaller number of levels, but each level requires a slower Fourier transform.
		Factors of 2 or 4 usually perform the best.
	window_size : integer
		The size of the next level in the number of pixels on the current layer. Lowering this
		increases performance in exchange for accuracy. Values smaller than 4-8 are not recommended.
	'''
	def __init__(self, input_grid, charge, lyot_stop=None, q=1024, scaling_factor=4, window_size=32):
		self.input_grid = input_grid
		pupil_diameter = input_grid.shape * input_grid.delta

		if hasattr(lyot_stop, 'forward') or lyot_stop is None:
			self.lyot_stop = lyot_stop
		else:
			self.lyot_stop = Apodizer(lyot_stop)

		levels = int(np.ceil(np.log(q / 2) / np.log(scaling_factor))) + 1
		qs = [2 * scaling_factor**i for i in range(levels)]
		num_airys = [input_grid.shape / 2]

		focal_grids = []
		self.focal_masks = []
		self.props = []

		for i in range(1, levels):
			num_airys.append(num_airys[i - 1] * window_size / (2 * qs[i - 1] * num_airys[i - 1]))

		for i in range(levels):
			q = qs[i]
			num_airy = num_airys[i]

			focal_grid = make_focal_grid(q, num_airy, pupil_diameter=pupil_diameter, reference_wavelength=1, focal_length=1)
			focal_mask = Field(np.exp(1j * charge * focal_grid.as_('polar').theta), focal_grid)

			focal_mask *= 1 - circular_aperture(1e-9)(focal_grid)

			if i != levels - 1:
				wx = windows.tukey(window_size, 1, False)
				wy = windows.tukey(window_size, 1, False)
				w = np.outer(wy, wx)

				w = np.pad(w, (focal_grid.shape - w.shape) // 2, 'constant').ravel()
				focal_mask *= 1 - w

			for j in range(i):
				fft = FastFourierTransform(focal_grids[j])
				mft = MatrixFourierTransform(focal_grid, fft.output_grid)

				focal_mask -= mft.backward(fft.forward(self.focal_masks[j]))

			if i == 0:
				prop = FourierFilter(input_grid, focal_mask, q)
			else:
				prop = FraunhoferPropagator(input_grid, focal_grid)

			focal_grids.append(focal_grid)
			self.focal_masks.append(focal_mask)
			self.props.append(prop)

	def forward(self, wavefront):
		'''Propagate a wavefront through the vortex coronagraph.

		Parameters
		----------
		wavefront : Wavefront
			The wavefront to propagate. This wavefront is expected to be
			in the pupil plane.

		Returns
		-------
		Wavefront
			The Lyot plane wavefront.
		'''
		wavelength = wavefront.wavelength
		wavefront.wavelength = 1

		for i, (mask, prop) in enumerate(zip(self.focal_masks, self.props)):
			if i == 0:
				lyot = Wavefront(prop.forward(wavefront.electric_field), input_stokes_vector=wavefront.input_stokes_vector)
			else:
				focal = prop(wavefront)
				focal.electric_field *= mask
				lyot.electric_field += prop.backward(focal).electric_field

		lyot.wavelength = wavelength
		wavefront.wavelength = wavelength

		if self.lyot_stop is not None:
			lyot = self.lyot_stop.forward(lyot)

		return lyot

	def backward(self, wavefront):
		'''Propagate backwards through the vortex coronagraph.

		This essentially is a forward propagation through a the same vortex
		coronagraph, but with the sign of the its charge flipped.

		Parameters
		----------
		wavefront : Wavefront
			The Lyot plane wavefront.

		Returns
		-------
		Wavefront
			The pupil-plane wavefront.
		'''
		if self.lyot_stop is not None:
			wavefront = self.lyot_stop.backward(wavefront)

		wavelength = wavefront.wavelength
		wavefront.wavelength = 1

		for i, (mask, prop) in enumerate(zip(self.focal_masks, self.props)):
			if i == 0:
				pup = Wavefront(prop.backward(wavefront.electric_field), input_stokes_vector=wavefront.input_stokes_vector)
			else:
				focal = prop(wavefront)
				focal.electric_field *= mask.conj()
				pup.electric_field += prop.backward(focal).electric_field

		pup.wavelength = wavelength
		wavefront.wavelength = wavelength

		return pup

class VectorVortexCoronagraph(AgnosticOpticalElement):
	'''An vector vortex coronagraph.

	This :class:`OpticalElement` simulations the propagation of light through
	a vector vortex in the focal plane. To resolve the singularity of this vortex
	phase plate, a multi-scale approach is made. Discretisation errors made at
	a certain level are corrected by the next level with finer sampling.

	Parameters
	----------
	charge : integer
		The charge of the vortex.
	lyot_stop : Field or OpticalElement
		The Lyot stop for the coronagraph. If it's a Field, it is converted to an
		OpticalElement for convenience. If this is None (default), then no Lyot stop is used.
	phase_retardation : scalar or function
		The phase retardation of the vector vortex plate, potentially as a
		function of wavelength. Changes of the phase retardation as a function
		of spatial position is not yet supported.
	q : scalar
		The minimum number of pixels per lambda/D. The number of levels in the multi-scale
		Fourier transforms will be chosen to reach at least this number of samples. The required
		q for a high-accuracy vortex coronagraph depends on the charge of the vortex. For charge 2,
		this can be as low as 32, but for charge 8 you need ~1024. Lower values give higher performance
		as a smaller number of levels is needed, but increases the sampling errors near the singularity.
		Charges not divisible by four require a much lower q. The default (q=1024) is conservative in
		most cases.
	scaling_factor : scalar
		The fractional increase in spatial frequency sampling per level. Larger scaling factors
		require a smaller number of levels, but each level requires a slower Fourier transform.
		Factors of 2 or 4 usually perform the best.
	window_size : integer
		The size of the next level in the number of pixels on the current layer. Lowering this
		increases performance in exchange for accuracy. Values smaller than 4-8 are not recommended.
	'''
	def __init__(self, charge, lyot_stop=None, phase_retardation=np.pi, q=1024, scaling_factor=4, window_size=32):
		self.charge = charge

		if hasattr(lyot_stop, 'forward') or lyot_stop is None:
			self.lyot_stop = lyot_stop
		else:
			self.lyot_stop = Apodizer(lyot_stop)

		self.phase_retardation = phase_retardation

		self.q = q
		self.scaling_factor = scaling_factor
		self.window_size = window_size

		AgnosticOpticalElement.__init__(self)

	def make_instance(self, instance_data, input_grid, output_grid, wavelength):
		pupil_diameter = input_grid.shape * input_grid.delta

		levels = int(np.ceil(np.log(self.q / 2) / np.log(self.scaling_factor))) + 1
		qs = [2 * self.scaling_factor**i for i in range(levels)]
		num_airys = [input_grid.shape / 2]

		focal_grids = []
		instance_data.props = []
		instance_data.jones_matrices = []

		for i in range(1, levels):
			num_airys.append(num_airys[i - 1] * self.window_size / (2 * qs[i - 1] * num_airys[i - 1]))

		for i in range(levels):
			q = qs[i]
			num_airy = num_airys[i]

			focal_grid = make_focal_grid(q, num_airy, pupil_diameter=pupil_diameter, reference_wavelength=1, focal_length=1)

			fast_axis_orientation = Field(self.charge / 2 * focal_grid.as_('polar').theta, focal_grid)
			retardance = self.evaluate_parameter(self.phase_retardation, input_grid, output_grid, wavelength)

			focal_mask_raw = LinearRetarder(retardance, fast_axis_orientation)
			jones_matrix = focal_mask_raw.jones_matrix

			jones_matrix *= 1 - circular_aperture(1e-9)(focal_grid)

			if i != levels - 1:
				wx = windows.tukey(self.window_size, 1, False)
				wy = windows.tukey(self.window_size, 1, False)
				w = np.outer(wy, wx)

				w = np.pad(w, (focal_grid.shape - w.shape) // 2, 'constant').ravel()
				jones_matrix *= 1 - w

			for j in range(i):
				fft = FastFourierTransform(focal_grids[j])
				mft = MatrixFourierTransform(focal_grid, fft.output_grid)

				jones_matrix -= mft.backward(fft.forward(instance_data.jones_matrices[j]))

			if i == 0:
				prop = FourierFilter(input_grid, jones_matrix, q)
			else:
				prop = FraunhoferPropagator(input_grid, focal_grid)

			focal_grids.append(focal_grid)
			instance_data.jones_matrices.append(jones_matrix)
			instance_data.props.append(prop)

	def get_input_grid(self, output_grid, wavelength):
		'''Get the input grid for a specified output grid and wavelength.

		This optical element only supports propagation to the same plane as
		its input.

		Parameters
		----------
		output_grid : Grid
			The output grid of the optical element.
		wavelength : scalar or None
			The wavelength of the outgoing light.

		Returns
		-------
		Grid
			The input grid corresponding to the output grid and wavelength combination.
		'''
		return output_grid

	def get_output_grid(self, input_grid, wavelength):
		'''Get the output grid for a specified input grid and wavelength.

		This optical element only supports propagation to the same plane as
		its input.

		Parameters
		----------
		input_grid : Grid
			The input grid of the optical element.
		wavelength : scalar or None
			The wavelength of the incoming light.

		Returns
		-------
		Grid
			The output grid corresponding to the input grid and wavelength combination.
		'''
		return input_grid

	@make_agnostic_forward
	def forward(self, instance_data, wavefront):
		'''Propagate a wavefront through the vortex coronagraph.

		Parameters
		----------
		wavefront : Wavefront
			The wavefront to propagate. This wavefront is expected to be
			in the pupil plane.

		Returns
		-------
		Wavefront
			The Lyot plane wavefront.
		'''
		wavelength = wavefront.wavelength
		wavefront.wavelength = 1

		for i, (jones_matrix, prop) in enumerate(zip(instance_data.jones_matrices, instance_data.props)):
			if i == 0:
				if not wavefront.is_polarized:
					wf = Wavefront(wavefront.electric_field, input_stokes_vector=(1, 0, 0, 0))
				else:
					wf = wavefront

				lyot = Wavefront(prop.forward(wf.electric_field), input_stokes_vector=wf.input_stokes_vector)
			else:
				focal = prop(wavefront)

				if not focal.is_polarized:
					focal = Wavefront(focal.electric_field, input_stokes_vector=(1, 0, 0, 0))

				focal.electric_field = field_dot(jones_matrix, focal.electric_field)
				if i == 0:
					lyot = prop.backward(focal)
				else:
					lyot.electric_field += prop.backward(focal).electric_field

		lyot.wavelength = wavelength
		wavefront.wavelength = wavelength

		if self.lyot_stop is not None:
			lyot = self.lyot_stop.forward(lyot)

		return lyot

	@make_agnostic_backward
	def backward(self, instance_data, wavefront):
		'''Propagate backwards through the vortex coronagraph.

		This essentially is a forward propagation through a the same vortex
		coronagraph, but with the sign of the its charge flipped.

		Parameters
		----------
		wavefront : Wavefront
			The Lyot plane wavefront.

		Returns
		-------
		Wavefront
			The pupil-plane wavefront.
		'''
		if self.lyot_stop is not None:
			wavefront = self.lyot_stop.backward(wavefront)

		wavelength = wavefront.wavelength
		wavefront.wavelength = 1

		for i, (jones_matrix, prop) in enumerate(zip(instance_data.jones_matrices, instance_data.props)):
			if i == 0:
				pup = Wavefront(prop.backward(wavefront.electric_field))
			else:
				focal = prop(wavefront)
				focal.electric_field = field_dot(jones_matrix.conj(), focal.electric_field)
				pup.electric_field += prop.backward(focal).electric_field

		pup.wavelength = wavelength
		wavefront.wavelength = wavelength

		return pup

def make_ravc_masks(central_obscuration, charge=2, pupil_diameter=1, lyot_undersize=0):
	'''Make field generators for the pupil and Lyot-stop masks for a
	ring apodized vortex coronagraph.

	The formulas were implemented according to [Mawet2013]_.

	.. [Mawet2013] Dimitri Mawet et al. 2013 "Ring-apodized vortex coronagraphs for obscured telescopes. I. Transmissive
		ring apodizers" The Astrophysical Journal Supplement Series 209.1 (2013): 7

	Parameters
	----------
	central_obscuration : scalar
		The diameter of the central obscuration.
	charge : integer
		The charge of the vortex coronagraph used.
	pupil_diameter : scalar
		The diameter of the pupil.
	lyot_undersize : scalar
		The fraction of the pupil diameter to which to undersize the Lyot stop.

	Returns
	-------
	pupil_mask : Field generator
		The complex transmission of the pupil mask.
	lyot_mask : Field generator
		The complex transmission of the Lyot-stop mask.
	'''
	R0 = central_obscuration / pupil_diameter

	if charge == 2:
		t1 = 1 - 0.25 * (R0**2 + R0 * np.sqrt(R0**2 + 8))
		R1 = R0 / np.sqrt(1 - t1)

		pupil1 = circular_aperture(pupil_diameter)
		pupil2 = circular_aperture(pupil_diameter * R1)
		co = circular_aperture(central_obscuration)
		pupil_mask = lambda grid: (pupil1(grid) * t1 + pupil2(grid) * (1 - t1)) * (1 - co(grid))

		lyot1 = circular_aperture(pupil_diameter * R1 + pupil_diameter * lyot_undersize)
		lyot2 = circular_aperture(pupil_diameter * (1 - lyot_undersize))
		lyot_stop = lambda grid: lyot2(grid) - lyot1(grid)
	elif charge == 4:
		R1 = np.sqrt(np.sqrt(R0**2 * (R0**2 + 4)) - 2*R0**2)
		R2 = np.sqrt(R1**2 + R0**2)
		t1 = 0
		t2 = (R1**2 - R0**2) / (R1**2 + R0**2)

		pupil1 = circular_aperture(pupil_diameter)
		pupil2 = circular_aperture(pupil_diameter * R1)
		pupil3 = circular_aperture(pupil_diameter * R2)
		co = circular_aperture(central_obscuration)

		pupil_mask = lambda grid: (pupil1(grid) * t2 + pupil3(grid) * (t1 - t2) + pupil2(grid) * (1 - t1)) * (1 - co(grid))

		lyot1 = circular_aperture(pupil_diameter * R2 + pupil_diameter * lyot_undersize)
		lyot2 = circular_aperture(pupil_diameter * (1 - lyot_undersize))
		lyot_stop = lambda grid: lyot2(grid) - lyot1(grid)
	else:
		raise NotImplementedError()

	return pupil_mask, lyot_stop

def get_ravc_planet_transmission(central_obscuration_ratio, charge=2):
	'''Get the planet transmission for a ring-apodized vortex coronagraph.

	The formulas were implemented according to [Mawet2013]_.

	.. [Mawet2013] Dimitri Mawet et al. 2013 "Ring-apodized vortex coronagraphs for obscured telescopes. I. Transmissive
		ring apodizers" The Astrophysical Journal Supplement Series 209.1 (2013): 7

	Parameters
	----------
	central_obscuration_ratio : scalar
		The ratio of the central obscuration diameter and the pupil diameter.
	charge : integer
		The charge of the vortex coronagraph used.

	Returns
	-------
	scalar
		The intensity transmission for a sufficiently off-axis point source
		for the ring-apodized vortex coronagraph. Point sources close to the vortex
		singularity will be lower in intensity.
	'''
	R0 = central_obscuration_ratio

	if charge == 2:
		t1_opt = 1 - 0.25 * (R0**2 + R0 * np.sqrt(R0**2 + 8))
		R1_opt = R0 / np.sqrt(1 - t1_opt)

		return t1_opt**2 * (1 - R1_opt**2) / (1 - (R0**2))
	elif charge == 4:
		R1 = np.sqrt(np.sqrt(R0**2 * (R0**2 + 4)) - 2*R0**2)
		R2 = np.sqrt(R1**2 + R0**2)
		t2 = (R1**2 - R0**2) / (R1**2 + R0**2)

		return t2**2 * (1 - R2**2) / (1 - R0**2)
	else:
		raise NotImplementedError()



from ..optics import Apodizer, OpticalElement
from ..propagation import FraunhoferPropagator

class LyotCoronagraph(OpticalElement):
	'''A Lyot coronagraph with a small focal-plane mask.

	The area outside of this focal-plane mask is assumed to be fully transmisive. The
	method for propagation is based on [Soummer2007]_.

	.. [Soummer2007] Soummer et al. 2007, "Fast computation of Lyot-style
		coronagraph propagation".

	Parameters
	----------
	input_grid : Grid
		The grid on which the incoming wavefront is defined.
	focal_plane_mask : Field or OpticalElement
		The (complex) transmission of the focal-plane mask. If this is an :class:`OpticalElement`,
		this will be used instead. This allows for more realistic implementations of focal-plane
		masks.
	lyot_stop : Field or OpticalElement or None
		The (complex) transmission of the Lyot stop. If this is an :class:`OpticalElement`,
		this will be used instead. This allows for more realistic implementations of Lyot stops.
	'''
	def __init__(self, input_grid, focal_plane_mask, lyot_stop=None):
		if hasattr(focal_plane_mask, 'input_grid'):
			# Focal plane mask is an optical element.
			grid = focal_plane_mask.input_grid
			self.focal_plane_mask = focal_plane_mask
		else:
			# Focal plane mask is a field.
			grid = focal_plane_mask.grid
			self.focal_plane_mask = Apodizer(focal_plane_mask)

		if lyot_stop is not None and not hasattr(lyot_stop, 'input_grid'):
			lyot_stop = Apodizer(lyot_stop)
		self.lyot_stop = lyot_stop

		self.prop = FraunhoferPropagator(input_grid, grid)

	def forward(self, wavefront):
		'''Propagate the wavefront through the Lyot coronagraph.

		Parameters
		----------
		wavefront : Wavefront
			The wavefront to propagate. This wavefront is assumed to be in the pupil plane.

		Returns
		-------
		Wavefront
			The Lyot-plane wavefront.
		'''
		wf_foc = self.prop.forward(wavefront)
		wf_foc.electric_field -= self.focal_plane_mask.forward(wf_foc).electric_field

		lyot = self.prop.backward(wf_foc)
		lyot.electric_field[:] = wavefront.electric_field - lyot.electric_field

		if self.lyot_stop is not None:
			lyot = self.lyot_stop.forward(lyot)

		return lyot

	def backward(self, wavefront):
		'''Propagate the wavefront from the Lyot plane to the pupil plane.

		Parameters
		----------
		wavefront : Wavefront
			The wavefront to propagate.

		Returns
		-------
		Wavefront
			The pupil-plane wavefront.
		'''
		if self.lyot_stop is not None:
			wf = self.lyot_stop.backward(wavefront)
		else:
			wf = wavefront

		wf_foc = self.prop.forward(wf)
		wf_foc.electric_field -= self.focal_plane_mask.backward(wf_foc).electric_field

		pup = self.prop.backward(wf_foc)
		pup.electric_field[:] = wf.electric_field - pup.electric_field

		return pup

class OccultedLyotCoronagraph(OpticalElement):
	'''A Lyot coronagraph with a focal-plane mask.

	The area outside of this focal-plane mask is assumed to be fully absorbing.

	Parameters
	----------
	input_grid : Grid
		The grid on which the incoming wavefront is defined.
	focal_plane_mask : Field or OpticalElement
		The (complex) transmission of the focal-plane mask. If this is an :class:`OpticalElement`,
		this will be used instead. This allows for more realistic implementations of focal-plane
		masks.
	lyot_stop : Field or OpticalElement or None
		The (complex) transmission of the Lyot stop. If this is an :class:`OpticalElement`,
		this will be used instead. This allows for more realistic implementations of Lyot stops.
	'''
	def __init__(self, input_grid, focal_plane_mask):
		if hasattr(focal_plane_mask, 'input_grid'):
			# Focal plane mask is an optical element.
			grid = focal_plane_mask.input_grid
			self.focal_plane_mask = focal_plane_mask
		else:
			# Focal plane mask is a field.
			grid = focal_plane_mask.grid
			self.focal_plane_mask = Apodizer(focal_plane_mask)

		self.prop = FraunhoferPropagator(input_grid, grid)

	def forward(self, wavefront):
		'''Propagate the wavefront through the Lyot coronagraph.

		Parameters
		----------
		wavefront : Wavefront
			The wavefront to propagate. This wavefront is assumed to be in the pupil plane.

		Returns
		-------
		Wavefront
			The Lyot-plane wavefront.
		'''
		return self.prop.backward(self.focal_plane_mask.forward(self.prop.forward(wavefront)))

	def backward(self, wavefront):
		'''Propagate the wavefront from the Lyot plane to the pupil plane.

		Parameters
		----------
		wavefront : Wavefront
			The wavefront to propagate.

		Returns
		-------
		Wavefront
			The pupil-plane wavefront.
		'''
		return self.prop.backward(self.focal_plane_mask.backward(self.prop.forward(wavefront)))



import numpy as np
from .apodization import SurfaceApodizer
from .optical_element import OpticalElement
from .surface_profiles import spherical_surface_sag, even_aspheric_surface_sag
from ..field import Field

def closest_points(point_grid, evaluated_grid):
	from scipy import spatial

	tree = spatial.KDTree(point_grid.points)
	d, i = tree.query(evaluated_grid.points)

	return Field(i, evaluated_grid), Field(d, evaluated_grid)

class MicroLensArray(OpticalElement):
	'''A parabolic micro-lens array.

	Parameters
	----------
	input_grid : Grid
		The grid on which the micro-lens array is evaluated.
	lenslet_grid : Grid
		A grid containing the lenslet positions.
	focal_length : scalar
		The focal length of the micro-lenses
	lenslet_shape : field generator
		The shape of a lenslet.
	'''
	def __init__(self, input_grid, lenslet_grid, focal_length, lenslet_shape=None):
		self.input_grid = input_grid
		self.focal_length = focal_length

		self.mla_grid = lenslet_grid

		if lenslet_shape is None:
			indices, distances = closest_points(lenslet_grid, input_grid)

			self.mla_index = indices
			self.mla_opd = (-1 / (2 * focal_length)) * distances**2
		else:
			self.mla_index = Field(-np.ones(self.input_grid.size), self.input_grid)
			self.mla_opd = Field(np.zeros(self.input_grid.size), self.input_grid)

			for i, (x,y) in enumerate(lenslet_grid.as_('cartesian').points):
				shifted_grid = input_grid.shifted((x,y))
				mask = lenslet_shape(shifted_grid) != 0

				self.mla_opd[mask] = (-1 / (2 * focal_length)) * (shifted_grid.x[mask]**2 + shifted_grid.y[mask]**2)
				self.mla_index[mask] = i

		self.mla_surface = SurfaceApodizer(self.mla_opd, 2)

	def forward(self, wavefront):
		return self.mla_surface.forward(wavefront)

	def backward(self, wavefront):
		return self.mla_surface.backward(wavefront)


class SphericalMicroLensArray(OpticalElement):
	'''An even asphere micro-lens array.

	Parameters
	----------
	input_grid : Grid
		The grid on which the micro-lens array is evaluated.
	lenslet_grid : Grid
		A grid containing the lenslet positions.
	radius_of_curvature : scalar
		The radius of curvature of the micro-lenses.
	lenslet_shape : field generator
		The shape of a lenslet.
	refractive_index : scalar or function
		The refractive index of the material.
	conic_constant : scalar
		The conic constant of the micro-lenses.
	aspheric_coefficients : array_like
		The aspheric coefficients of the micro-lenses.
	'''
	def __init__(self, input_grid, lenslet_grid, radius_of_curvature, lenslet_shape, refractive_index=1.5):

		self.input_grid = input_grid
		self.mla_grid = lenslet_grid
		self.n = refractive_index
		self.radius_of_curvature = radius_of_curvature

		self.mla_index = -self.input_grid.ones()
		self.mla_opd = self.input_grid.zeros()
		self.surface_sag = spherical_surface_sag(radius_of_curvature)

		for i, (x,y) in enumerate(lenslet_grid.as_('cartesian').points):
			shifted_grid = input_grid.shifted((x,y))
			mask = lenslet_shape(shifted_grid) != 0
			subset_grid = shifted_grid.subset(mask)

			# Check if the micro-lens is within the field of view.
			if np.count_nonzero(mask) > 0:
				self.mla_index[mask] = i
				self.mla_opd[mask] += self.surface_sag(subset_grid)

		self.mla_surface = SurfaceApodizer(self.mla_opd, refractive_index)

	def forward(self, wavefront):
		return self.mla_surface.forward(wavefront)

	def backward(self, wavefront):
		return self.mla_surface.backward(wavefront)

class EvenAsphereMicroLensArray(OpticalElement):
	'''An even asphere micro-lens array.

	Parameters
	----------
	input_grid : Grid
		The grid on which the micro-lens array is evaluated.
	lenslet_grid : Grid
		A grid containing the lenslet positions.
	radius_of_curvature : scalar
		The radius of curvature of the micro-lenses.
	lenslet_shape : field generator
		The shape of a lenslet.
	refractive_index : scalar or function
		The refractive index of the material.
	conic_constant : scalar
		The conic constant of the micro-lenses.
	aspheric_coefficients : array_like
		The aspheric coefficients of the micro-lenses.
	'''
	def __init__(self, input_grid, lenslet_grid, radius_of_curvature, lenslet_shape, refractive_index=1.5, conic_constant=0, aspheric_coefficients=None):
		self.input_grid = input_grid
		self.mla_grid = lenslet_grid
		self.n = refractive_index
		self.radius_of_curvature = radius_of_curvature
		self.conic_constant = conic_constant
		if aspheric_coefficients is None:
			self.aspheric_coefficients = []
		else:
			self.aspheric_coefficients = aspheric_coefficients

		self.mla_index = -self.input_grid.ones()
		self.mla_opd = self.input_grid.zeros()
		self.surface_sag = even_aspheric_surface_sag(radius_of_curvature, conic_constant, aspheric_coefficients)

		for i, (x,y) in enumerate(lenslet_grid.as_('cartesian').points):
			shifted_grid = input_grid.shifted((x,y))
			mask = lenslet_shape(shifted_grid) != 0
			subset_grid = shifted_grid.subset(mask)

			# Check if the micro-lens is within the field of view.
			if np.count_nonzero(mask) > 0:
				self.mla_index[mask] = i
				self.mla_opd[mask] += self.surface_sag(subset_grid)

		self.mla_surface = SurfaceApodizer(self.mla_opd, refractive_index)

	def forward(self, wavefront):
		return self.mla_surface.forward(wavefront)

	def backward(self, wavefront):
		return self.mla_surface.backward(wavefront)



import numpy as np

from ..optics import Wavefront, AgnosticOpticalElement, make_agnostic_forward, make_agnostic_backward
from ..field import Field, evaluate_supersampled
from ..fourier import FastFourierTransform, make_fft_grid, FourierFilter

class AngularSpectrumPropagator(AgnosticOpticalElement):
	'''The monochromatic angular spectrum propagator for scalar fields.

	The scalar Angular Spectrum propagator is implemented as described by
	[McLeod2014]_. The propagation of an electric field can be described as a transfer
	function in frequency space. The transfer function is taken from
	equation 9 of [McLeod2014]_, and the related impulse response is taken from
	equation 6 of [McLeod2014]_.

	.. [McLeod2014] Robert R. McLeod and Kelvin H. Wagner 2014, "Vector Fourier optics of
		anisotropic materials," Adv. Opt. Photon. 6, 368-412 (2014)

	Parameters
	----------
	input_grid : anything
		This argument is ignored. The input grid is taken from the incoming wavefront.
	distance : scalar
		The distance to propagate
	num_oversampling : int
		The number of times the transfer function is oversampled. Default is 2.
	wavelength : scalar
		The wavelength of the wavefront.
	refractive_index : scalar
		The refractive index of the medium that the wavefront is propagating in.

	Raises
	------
	ValueError
		If the `input_grid` is not regular and Cartesian.
	'''
	def __init__(self, input_grid, distance, num_oversampling=2, refractive_index=1):
		self._distance = distance

		self._num_oversampling = num_oversampling
		self._refractive_index = refractive_index

		AgnosticOpticalElement.__init__(self, grid_dependent=True, wavelength_dependent=True)

	def make_instance(self, instance_data, input_grid, output_grid, wavelength):
		if not input_grid.is_regular or not input_grid.is_('cartesian'):
			raise ValueError('The input grid must be a regular, Cartesian grid.')

		k = 2 * np.pi / wavelength * self.evaluate_parameter(self.refractive_index, input_grid, output_grid, wavelength)
		L_max = np.max(input_grid.dims * input_grid.delta)

		if np.any(input_grid.delta < wavelength * abs(self.distance) / L_max):
			def transfer_function(fourier_grid):
				enlarged_grid = make_fft_grid(fourier_grid)
				fft_upscale = FastFourierTransform(enlarged_grid)

				def impulse_response(grid):
					r_squared = grid.x**2 + grid.y**2 + self.distance**2
					r = np.sqrt(r_squared)
					cos_theta = self.distance / r

					return Field(cos_theta / (2 * np.pi) * np.exp(1j * k * r) * (1 / r_squared - 1j * k / r), grid)

				impulse_response = evaluate_supersampled(impulse_response, enlarged_grid, self.num_oversampling)

				return fft_upscale.forward(impulse_response)
		else:
			def transfer_function_native(fourier_grid):
				k_squared = fourier_grid.as_('polar').r**2
				k_z = np.sqrt(k**2 - k_squared + 0j)

				return Field(np.exp(1j * k_z * self.distance), fourier_grid)

			def transfer_function(fourier_grid):
				return evaluate_supersampled(transfer_function_native, fourier_grid, self.num_oversampling)

		instance_data.fourier_filter = FourierFilter(input_grid, transfer_function, q=2)

	@property
	def distance(self):
		return self._distance

	@distance.setter
	def distance(self, distance):
		self._distance = distance

		self.clear_cache()

	@property
	def num_oversampling(self):
		return self._num_oversampling

	@num_oversampling.setter
	def num_oversampling(self, num_oversampling):
		self._num_oversampling = num_oversampling

		self.clear_cache()

	@property
	def refractive_index(self):
		return self._refractive_index

	@refractive_index.setter
	def refractive_index(self, refractive_index):
		self._refractive_index = refractive_index

		self.clear_cache()

	def get_input_grid(self, output_grid, wavelength):
		return output_grid

	def get_output_grid(self, input_grid, wavelength):
		return input_grid

	@make_agnostic_forward
	def forward(self, instance_data, wavefront):
		'''Propagate a wavefront forward by a certain distance.

		Parameters
		----------
		wavefront : Wavefront
			The incoming wavefront.

		Returns
		-------
		Wavefront
			The wavefront after the propagation.
		'''
		filtered = instance_data.fourier_filter.forward(wavefront.electric_field)

		return Wavefront(filtered, wavefront.wavelength, wavefront.input_stokes_vector)

	@make_agnostic_backward
	def backward(self, instance_data, wavefront):
		'''Propagate a wavefront backward by a certain distance.

		Parameters
		----------
		wavefront : Wavefront
			The incoming wavefront.

		Returns
		-------
		Wavefront
			The wavefront after the propagation.
		'''
		filtered = instance_data.fourier_filter.backward(wavefront.electric_field)

		return Wavefront(filtered, wavefront.wavelength, wavefront.input_stokes_vector)



import numpy as np

def get_strehl_from_focal(img, ref_img):
	'''Get the Strehl ratio from a focal-plane image.

	Parameters
	----------
	img : Field or array_like
		The focal-plane image.
	ref_img : Field or array_like
		The reference focal-plane image without aberrations.

	Returns
	-------
	scalar
		The Strehl ratio.
	'''
	return img[np.argmax(ref_img)] / ref_img.max()

def get_strehl_from_pupil(aperture, ref_aperture):
	'''Get the Strehl ratio from a pupil-plane electric field.

	Parameters
	----------
	aperture : Field or array_like
		The pupil-plane electric field.
	ref_aperture : Field or array_like
		The reference pupil-plane electric field without aberrations.

	Returns
	-------
	scalar
		The Strehl ratio.
	'''
	return np.abs(np.sum(aperture) / np.sum(ref_aperture))**2

def get_mean_intensity_in_roi(img, mask):
	'''Get the mean intensity in a masked region of interest.

	Parameters
	----------
	img : Field or array_like
		The focal-plane image.
	mask : Field or array_like
		A binary array describing the region of interest.

	Returns
	-------
	scalar
		The mean intensity in the region of interest.
	'''
	return np.mean(img[mask])

def get_mean_raw_contrast(img, mask, ref_img):
	'''Get the mean raw contrast in a masked region of interest.

	Parameters
	----------
	img : Field or array_like
		The focal-plane image.
	mask : Field or array_like
		A binary array describing the region of interest.
	img_ref : Field or array_like
		A reference focal-plane image without aberrations. This is used
		to determine the Strehl ratio.

	Returns
	-------
	scalar
		The mean raw contrast in the region of interest.
	'''
	mean_intensity = get_mean_intensity_in_roi(img, mask)
	strehl = get_strehl_from_focal(img, ref_img)

	return mean_intensity / strehl



from .optical_element import AgnosticOpticalElement, make_agnostic_forward, make_agnostic_backward

import numpy as np

class Magnifier(AgnosticOpticalElement):
	def __init__(self, magnification):
		''' An ideal magnifier.

		This class magnifies a wavefront in an energy conserving manner.

		Parameters
		----------
		magnification : scalar or function of wavelength
			The magnification of the system.
		'''
		self._magnification = magnification

		AgnosticOpticalElement.__init__(self, False, True)

	def make_instance(self, instance_data, input_grid, output_grid, wavelength):
		instance_data.magnification = self.evaluate_parameter(self._magnification, input_grid, output_grid, wavelength) * np.ones(2)

	@property
	def magnification(self):
		return self._magnification

	@magnification.setter
	def magnification(self, magnification):
		self._magnification = magnification

		self.clear_cache()

	def get_input_grid(self, output_grid, wavelength):
		magnification = self.evaluate_parameter(self._magnification, None, None, wavelength) * np.ones(2)

		return output_grid.scaled(1.0 / magnification)

	def get_output_grid(self, input_grid, wavelength):
		magnification = self.evaluate_parameter(self._magnification, input_grid, None, wavelength) * np.ones(2)

		return input_grid.scaled(magnification)

	@make_agnostic_forward
	def forward(self, instance_data, wavefront):
		wf = wavefront.copy()

		wf.electric_field.grid = wf.electric_field.grid.scaled(instance_data.magnification)
		wf.electric_field /= np.sqrt(np.prod(instance_data.magnification))

		return wf

	@make_agnostic_backward
	def backward(self, instance_data, wavefront):
		wf = wavefront.copy()

		wf.electric_field.grid = wf.electric_field.grid.scaled(1.0 / instance_data.magnification)
		wf.electric_field *= np.sqrt(np.prod(instance_data.magnification))

		return wf



import numpy as np
from .grid import Grid
from .coordinates import UnstructuredCoords

from functools import reduce
import operator

def _prod(iterable):
	return reduce(operator.mul, iterable, 1)

def _get_rotation_matrix(ndim, angle, axis=None):
	if ndim == 1:
		raise ValueError('Rotation of a one-dimensional grid is not possible.')
	elif ndim > 3:
		raise NotImplementedError()

	if ndim == 2:
		return np.array([[np.cos(angle), -np.sin(angle)], [np.sin(angle), np.cos(angle)]])
	elif ndim == 3:
		if axis is None:
			raise ValueError('An axis must be supplied when rotating a three-dimensional grid.')

		axis = np.array(axis).astype('float')
		if np.all(np.array(axis.shape) != (3,)):
			raise ValueError('The axis must be a 3-vector.')
		axis /= np.sqrt(axis.dot(axis))

		K = np.array([[0, -axis[2], axis[1]],[axis[2], 0, -axis[0]], [-axis[1], axis[0], 0]])
		return np.eye(3) + np.sin(angle) * K + (1 - np.cos(angle)) * K.dot(K)

class CartesianGrid(Grid):
	'''A grid representing a N-dimensional Cartesian coordinate system.
	'''
	_coordinate_system = 'cartesian'

	@property
	def x(self):
		'''The x-coordinate (dimension 0).
		'''
		return self.coords[0]

	@property
	def y(self):
		'''The y-coordinate (dimension 1).
		'''
		return self.coords[1]

	@property
	def z(self):
		'''The z-coordinate (dimension 2).
		'''
		return self.coords[2]

	@property
	def w(self):
		'''The w-coordinate (dimension 3).
		'''
		return self.coords[3]

	def scale(self, scale):
		'''Scale the grid in-place.

		Parameters
		----------
		scale : array_like or scalar
			The factor with which to scale the grid.

		Returns
		-------
		Grid
			Itself to allow for chaining these transformations.
		'''
		if np.isscalar(scale):
			self.weights *= np.abs(scale)**self.ndim
		else:
			self.weights *= np.prod(np.abs(scale))

		self.coords *= scale

		return self

	def shift(self, shift):
		'''Shift the grid in-place.

		Parameters
		----------
		shift : array_like
			The amount with which to shift the grid.

		Returns
		-------
		Grid
			Itself to allow for chaining these transformations.
		'''
		self.coords += shift
		return self

	def rotate(self, angle, axis=None):
		'''Rotate the grid in-place.

		.. caution::
			All structure in the coordinates will be destroyed.

		Parameters
		----------
		angle : scalar
			The angle in radians.
		axis : ndarray or None
			The axis of rotation. For two-dimensional grids, it is ignored. For
			three-dimensional grids it is required.

		Returns
		-------
		Grid
			Itself to allow for chaining these transformations.
		'''
		R = _get_rotation_matrix(self.ndim, angle, axis)

		coords = np.einsum('ik,kn->in', R, np.array(self.coords))
		self.coords = UnstructuredCoords(coords)
		return self

	def rotated(self, angle, axis=None):
		'''A rotated copy of this grid.

		Parameters
		----------
		angle : scalar
			The angle in radians.
		axis : ndarray or None
			The axis of rotation. For two-dimensional grids, it is ignored. For
			three-dimensional grids it is required.

		Returns
		-------
		Grid
			The rotated grid.
		'''
		R = _get_rotation_matrix(self.ndim, angle, axis)

		coords = np.einsum('ik,kn->in', R, np.array(self.coords))
		return CartesianGrid(UnstructuredCoords(coords))

	@staticmethod
	def _get_automatic_weights(coords):
		if coords.is_regular:
			return np.prod(coords.delta)
		elif coords.is_separated:
			weights = []
			for i in range(len(coords)):
				x = coords.separated_coords[i]
				w = (x[2:] - x[:-2]) / 2.
				w = np.concatenate(([x[1] - x[0]], w, [x[-1] - x[-2]]))
				weights.append(w)

			return _prod(np.ix_(*weights[::-1])).ravel()

Grid._add_coordinate_system('cartesian', CartesianGrid)



import numpy as np
import scipy.sparse

from ..field import Field, Grid

class ModeBasis(object):
	'''A linear basis of modes.

	Modes can be stored both in a dense or sparse format. A sparse mode basis reduces memory usage when modes
	are localized and contain a lot of zeros. Projection onto and linear combinations of modes are done seamlessly,
	regardless of the sparsity of the mode basis. If a grid is available, linear combinations automatically become
	Fields.

	You can create your own mode bases by supplying a list of modes, or a transformation matrix that transforms from
	mode coefficients to the mode. You can also use a number of built-in mode bases in HCIPy.

	Parameters
	----------
	transformation_matrix : array_like or list of array_like
		The transformation matrix of the mode basis or a list of modes.
	grid : Grid or None
		The grid on which the modes are defined.
	'''
	def __init__(self, transformation_matrix, grid=None):
		if scipy.sparse.issparse(transformation_matrix):
			sparse = True
			is_list = False
		elif scipy.sparse.issparse(transformation_matrix[0]):
			sparse = np.all([transformation_matrix[i].shape[0] == 1 for i in range(len(transformation_matrix))])
			is_list = True
		else:
			sparse = False
			is_list = isinstance(transformation_matrix, (list, tuple))

		if sparse:
			if is_list:
				self._modes = scipy.sparse.vstack(transformation_matrix, format='csr')
				self._transformation_matrix = self._modes.T.tocsc()
			else:
				self._modes = transformation_matrix.T.tocsr()
				self._transformation_matrix = transformation_matrix.tocsc()
		else:
			if is_list:
				self._transformation_matrix = np.stack(transformation_matrix, axis=-1)
			else:
				self._transformation_matrix = transformation_matrix

		if grid is not None:
			self.grid = grid
		elif hasattr(transformation_matrix[0], 'grid'):
			self.grid = transformation_matrix[0].grid
		else:
			self.grid = None

	@classmethod
	def from_dict(cls, tree):
		'''Make a ModeBasis from a dictionary, previously created by `to_dict()`.

		Parameters
		----------
		tree : dictionary
			The dictionary from which to make a new ModeBasis object.

		Returns
		-------
		ModeBasis
			The created object.

		Raises
		------
		ValueError
			If the dictionary is not formatted correctly.
		'''
		if isinstance(tree['transformation_matrix'], dict):
			data = np.array(tree['transformation_matrix']['data'])
			indices = np.array(tree['transformation_matrix']['indices'])
			indptr = np.array(tree['transformation_matrix']['indptr'])
			shape = tree['transformation_matrix']['shape']

			transformation_matrix = scipy.sparse.csc_matrix((data, indices, indptr), shape=shape)
		else:
			transformation_matrix = np.array(tree['transformation_matrix'])

		if 'grid' in tree:
			mode_basis = cls(transformation_matrix, Grid.from_dict(tree['grid']))
		else:
			mode_basis = cls(transformation_matrix)

		if tree['is_sparse']:
			return mode_basis.to_sparse()
		else:
			return mode_basis.to_dense()

	def to_dict(self):
		'''Convert the object to a dictionary for serialization.

		Returns
		-------
		dictionary
			The created dictionary.
		'''
		if self.is_sparse:
			transformation_matrix = {
				'data': self._transformation_matrix.data,
				'indices': self._transformation_matrix.indices,
				'indptr': self._transformation_matrix.indptr,
				'shape': list(self._transformation_matrix.shape)
			}
		else:
			transformation_matrix = self.transformation_matrix

		tree = {
			'grid': self.grid.to_dict(),
			'transformation_matrix': transformation_matrix,
			'is_sparse': self.is_sparse
		}

		return tree

	@property
	def is_sparse(self):
		'''If the mode basis is sparse.
		'''
		return scipy.sparse.issparse(self._transformation_matrix)

	@property
	def is_dense(self):
		'''If the mode basis is dense.
		'''
		return not self.is_sparse

	def to_sparse(self, copy=False):
		'''Convert the mode basis to a sparse mode basis.

		Parameters
		----------
		copy : boolean
			Whether to force a copy or not. A copy is always made if
			the current ModeBasis is not sparse.

		Returns
		-------
		ModeBasis
			The sparsified ModeBasis.

		Raises
		------
			TypeError
				If this ModeBasis cannot be sparsified.
		'''
		if self.is_sparse:
			if copy:
				return ModeBasis(self._transformation_matrix.copy(), self.grid)
			else:
				return self
		else:
			if self._transformation_matrix.ndim != 2:
				raise TypeError('Cannot sparsify a mode basis of tensor fields')

			T = scipy.sparse.csc_matrix(self._transformation_matrix)
			T.eliminate_zeros()
			return ModeBasis(T, self.grid)

	def to_dense(self, copy=False):
		'''Convert the mode basis to a dense mode basis.

		Parameters
		----------
		copy : boolean
			Whether to force a copy or not. A copy is always made if
			the current ModeBasis is not dense.

		Returns
		-------
		ModeBasis
			The densified ModeBasis.
		'''
		if self.is_dense:
			if copy:
				return ModeBasis(self._transformation_matrix.copy(), self.grid)
			else:
				return self
		else:
			T = np.asarray(self._transformation_matrix.todense())
			return ModeBasis(T, self.grid)

	@property
	def transformation_matrix(self):
		'''The transformation matrix of this mode basis.
		'''
		return self._transformation_matrix

	@transformation_matrix.setter
	def transformation_matrix(self, transformation_matrix):
		self._transformation_matrix = transformation_matrix

	def coefficients_for(self, b, dampening_factor=0):
		r'''Calculate the coefficients on this mode basis in a least squares fashion.

		The vector `b` is projection onto the mode basis in a least squares fashion. This
		means that the const function

		.. math:: J(c) = |b - A x|^2_2 + |\lambda x|^2_2

		is minimized, where :math:`x` are the coefficients, and :math:`\lambda` is the
		dampening factor.

		If this projection needs to be done repeatedly, you may be better off calculating
		the inverse of the transformation matrix directly and left-multiplying that with
		your vector, rather than using a least squares estimation every time.

		Parameters
		----------
		b : array_like or Field
			The vector for which to calculate the coefficients.
		dampening_factor : scalar
			The Tikhonov dampening factor used for the least squares procedure.

		Returns
		-------
		array_like
			The coefficients that correspond to the vector `b`.
		'''
		if self.is_sparse or dampening_factor != 0:
			x, istop, itn, normr, normar, norma, conda, normx = scipy.sparse.linalg.lsmr(self._transformation_matrix, b, damp=dampening_factor)
			return x
		else:
			x, residuals, rank, s = np.linalg.lstsq(self._transformation_matrix, b)
			return x

	def linear_combination(self, coefficients):
		'''Calculate a linear combination using this mode basis.

		Parameters
		----------
		coefficients : array_like or list
			The coefficients of the linear combinations.

		Returns
		-------
		array_like or Field
			The calculated linear combination.
		'''
		y = self._transformation_matrix.dot(coefficients)

		if self.grid is None:
			return y
		else:
			return Field(y, self.grid)

	@property
	def orthogonalized(self):
		'''Get an orthogonalized version of this ModeBasis.

		The resulting ModeBasis spans the same vector space, but each mode is orthogonal to
		all others. In general the resulting `ModeBasis` is dense, so no distinction is made
		between sparse and dense mode bases in this function. This function will always return
		a dense mode basis.

		Returns
		-------
		ModeBasis
			A mode basis with orthogonalized modes.

		Raises
		------
		NotImplementedError
			If the mode basis is a mode basis containing non-scalar fields.
		'''
		if self._transformation_matrix.ndim != 2:
			raise NotImplementedError('The mode basis contains non-scalar fields; orthogonalization is not implemented for these.')

		q, r = np.linalg.qr(self._transformation_matrix)
		return ModeBasis(q, self.grid)

	def __getitem__(self, item):
		'''Get the `item`-th mode in the `ModeBasis`.

		Parameters
		----------
		item : int or slice or array_like
			The index/indices of the mode(s).

		Returns
		-------
		Field or array_like or ModeBasis
			The `item`-th mode in the `ModeBasis`.
		'''
		T = self._transformation_matrix[..., item]

		return_mode_basis = False
		if self.is_sparse:
			if T.shape[-1] != 1:
				return_mode_basis = True
		if self.is_dense:
			if T.ndim == self._transformation_matrix.ndim:
				return_mode_basis = True

		if return_mode_basis:
			# We are returning multiple modes; put these in a ModeBasis.
			return ModeBasis(T, self.grid)
		else:
			# We are returning a single mode; return just this.
			if self.is_sparse:
				T = T.toarray()[...,0]

			if self.grid is None:
				return T
			else:
				return Field(T, self.grid)

	def __len__(self):
		'''The number of modes in the `ModeBasis`.

		Returns
		-------
		int
			The number of modes in the `ModeBasis`.
		'''
		return self.transformation_matrix.shape[-1]

	@property
	def num_modes(self):
		'''The number of modes in the `ModeBasis`.
		'''
		return len(self)

	def append(self, mode):
		'''Append `mode` to this mode basis.

		Parameters
		----------
		mode : array_like or Field
			The mode to add to the ModeBasis
		'''
		if self.is_sparse:
			self._transformation_matrix = scipy.sparse.hstack((self._transformation_matrix, mode), 'csc')
		else:
			self._transformation_matrix = np.concatenate((self._transformation_matrix, [mode]), axis=-1)

	def extend(self, modes):
		'''Extend the mode basis with `modes`.

		Parameters
		----------
		modes : list or array_like or ModeBasis
			The modes to add to the ModeBasis.
		'''
		if isinstance(modes, ModeBasis):
			modes = modes.transformation_matrix

		if self.is_sparse:
			self._transformation_matrix = scipy.sparse.hstack((self._transformation_matrix, modes), 'csc')
		else:
			# TODO: worry about modes being in a list instead of transformation matrix or mode basis
			self._transformation_matrix = np.concatenate((self._transformation_matrix, modes), axis=-1)

	def __add__(self, mode_basis):
		'''Merge two mode bases into one.

		Parameters
		----------
		mode_basis : ModeBasis
			The ModeBasis to add.

		Returns
		-------
		ModeBasis
			The newly created `ModeBasis`.
		'''
		if self.is_sparse or mode_basis.is_sparse:
			transformation_matrix = scipy.sparse.hstack((self._transformation_matrix, mode_basis.transformation_matrix), 'csc')
		else:
			transformation_matrix = np.concatenate((self._transformation_matrix, mode_basis.transformation_matrix), axis=-1)
		return ModeBasis(transformation_matrix, self.grid)



import copy
import numpy as np
import numexpr as ne

from ..field import Field, field_dot, field_kron

_U_matrix = 1 / np.sqrt(2) * np.array([
	[1, 0, 0, 1],
	[1, 0, 0, -1],
	[0, 1, 1, 0],
	[0, 1j, -1j, 0]])

# TODO Should add a pilot Gaussian beam with each Wavefront

class Wavefront(object):
	'''A physical wavefront in an optical system.

	This represents the state of light to be propagated through the
	optical system. It can be both an electric field in the scalar
	approximation (ie. scalar wavefront propgation), a fully polarized
	wavefront, represented by a Field of Jones vectors, and a potentially
	partially-polarized wavefront, represented by two Jones vector fields
	and the Stokes vector corresponding to the Jones vectors (1,0) and (0,1).

	Parameters
	----------
	electric_field : Field
		The electric field, either scalar, vectorial or 2-tensorial.
	wavelength : scalar
		The wavelength of the wavefront.
	input_stokes_vector : ndarray or None
		If a Stokes vector (I, Q, U, V) is given, a partially-polarized
		wavefront is initialized. If `electric_field` is scalar, it will
		be transformed into a tensor field with the correct Jones states.
		If a tensor-field is given as the `electric_field`, the electric
		field will be interpreted as the Jones matrix modifying the input
		Stokes vector.

	Raises
	------
	ValueError
		When a Stokes vector is supplied but a vector field is given as
		electric field, or when an input Stokes vector is not supplied,
		but a 2-tensor field is given as electric field.
	'''
	def __init__(self, electric_field, wavelength=1, input_stokes_vector=None):
		if input_stokes_vector is not None:
			if electric_field.tensor_order not in [0, 2]:
				raise ValueError('When supplying a Stokes vector, the electric field must be either a scalar or 2-tensor field.')

			if electric_field.is_scalar_field:
				self._electric_field = electric_field[np.newaxis, np.newaxis, :].astype('complex') * np.eye(2)[..., np.newaxis]
			else:
				self._electric_field = electric_field.astype('complex')

			self._input_stokes_vector = np.array(input_stokes_vector)
		else:
			self._electric_field = electric_field.astype('complex')
			self._input_stokes_vector = None

			if electric_field.tensor_order == 2:
				raise ValueError('When supplying a 2-tensor field as electric field, an input Stokes vector is required.')

		self.wavelength = wavelength

	def copy(self):
		'''Make a copy of the wavefront.
		'''
		return copy.deepcopy(self)

	@property
	def electric_field(self):
		'''The electric field as function of 2D position on the plane.
		'''
		return self._electric_field

	@electric_field.setter
	def electric_field(self, U):
		if hasattr(U, 'grid'):
			self._electric_field = U.astype('complex')
		else:
			if len(U) == 2:
				self._electric_field = Field(U[0].astype('complex'), U[1])
			else:
				raise ValueError("Electric field requires an accompanying grid.")

	@property
	def input_stokes_vector(self):
		'''The Stokes vector corresponding to the Jones vectors (1,0) and (0,1).
		'''
		return self._input_stokes_vector

	@property
	def wavenumber(self):
		'''The wavenumber of the light.
		'''
		return 2 * np.pi / self.wavelength

	@wavenumber.setter
	def wavenumber(self, wavenumber):
		self.wavelength = 2 * np.pi / wavenumber

	@property
	def grid(self):
		'''The grid on which the electric field is defined.
		'''
		return self.electric_field.grid

	@property
	def I(self):
		'''The I-component of the Stokes vector as function of 2D position
		in the plane.
		'''
		if self.is_scalar:
			# This is a scaler field.
			return np.abs(self.electric_field)**2
		elif self.is_partially_polarized:
			# This is a tensor field.
			x = self._electric_field[0, 0, :]
			y = self._electric_field[0, 1, :]
			z = self._electric_field[1, 0, :]
			w = self._electric_field[1, 1, :]

			a, b, c, d = self._input_stokes_vector

			# NumExpr is not smart enough to have abs(x) be real, so we have to take the real
			# part manually. This may change in later releases of NumExpr.
			M11 = 'real(abs(x))**2 + real(abs(y))**2 + real(abs(z))**2 + real(abs(w))**2'
			M12 = 'real(abs(x))**2 - real(abs(y))**2 + real(abs(z))**2 - real(abs(w))**2'
			M13 = '2 * (real(x) * real(y) + imag(x) * imag(y) + real(z) * real(w) + imag(z) * imag(w))'
			M14 = '2 * (-real(x) * imag(y) + imag(x) * real(y) - real(z) * imag(w) + imag(z) * real(w))'

			res = '0.5 * ((' + M11 + ') * a + (' + M12 + ') * b + (' + M13 + ') * c +  (' + M14 + ') * d)'
			return Field(ne.evaluate(res), self.electric_field.grid)
		else:
			# This is a vector field.
			return np.sum(np.abs(self.electric_field)**2, axis=0)

	@property
	def Q(self):
		'''The Q-component of the Stokes vector as function of 2D position
		in the plane.
		'''
		if self.is_scalar:
			# This is a scaler field.
			return self.grid.zeros()
		elif self.is_partially_polarized:
			# This is a tensor field.
			x = self._electric_field[0, 0, :]
			y = self._electric_field[0, 1, :]
			z = self._electric_field[1, 0, :]
			w = self._electric_field[1, 1, :]

			a, b, c, d = self._input_stokes_vector

			M21 = 'real(abs(x))**2 + real(abs(y))**2 - real(abs(z))**2 - real(abs(w))**2'
			M22 = 'real(abs(x))**2 - real(abs(y))**2 - real(abs(z))**2 + real(abs(w))**2'
			M23 = '2 * (real(x) * real(y) + imag(x) * imag(y) - real(z) * real(w) - imag(z) * imag(w))'
			M24 = '2 * (-real(x) * imag(y) + imag(x) * real(y) + real(z) * imag(w) - imag(z) * real(w))'

			res = '0.5 * ((' + M21 + ') * a + (' + M22 + ') * b + (' + M23 + ') * c +  (' + M24 + ') * d)'
			return Field(ne.evaluate(res), self.electric_field.grid)
		else:
			# This is a vector field.
			return np.abs(self.electric_field[0,:])**2 - np.abs(self.electric_field[1,:])**2

	@property
	def U(self):
		'''The U-component of the Stokes vector as function of 2D position
		in the plane.
		'''
		if self.is_scalar:
			# This is a scaler field.
			return self.grid.zeros()
		elif self.is_partially_polarized:
			# This is a tensor field.
			x = self._electric_field[0, 0, :]
			y = self._electric_field[0, 1, :]
			z = self._electric_field[1, 0, :]
			w = self._electric_field[1, 1, :]

			a, b, c, d = self._input_stokes_vector

			M31 = '2 * (real(x) * real(z) + imag(x) * imag(z) + real(y) * real(w) + imag(y) * imag(w))'
			M32 = '2 * (real(x) * real(z) + imag(x) * imag(z) - real(y) * real(w) - imag(y) * imag(w))'
			M33 = '2 * (real(x) * real(w) + imag(x) * imag(w) + real(y) * real(z) + imag(y) * imag(z))'
			M34 = '2 * (imag(x) * real(w) - real(x) * imag(w) + real(y) * imag(z) - imag(y) * real(z))'

			res = '0.5 * ((' + M31 + ') * a + (' + M32 + ') * b + (' + M33 + ') * c +  (' + M34 + ') * d)'
			return Field(ne.evaluate(res), self.electric_field.grid)
		else:
			# This is a vector field.
			return 2 * np.real(self.electric_field[0,:] * self.electric_field[1,:].conj())

	@property
	def V(self):
		'''The V-component of the Stokes vector as function of 2D position
		in the plane.
		'''
		if self.is_scalar:
			# This is a scaler field.
			return self.grid.zeros()
		elif self.is_partially_polarized:
			# This is a tensor field.
			x = self._electric_field[0, 0, :]
			y = self._electric_field[0, 1, :]
			z = self._electric_field[1, 0, :]
			w = self._electric_field[1, 1, :]

			a, b, c, d = self._input_stokes_vector

			M41 = '2 * (real(x) * imag(z) - imag(x) * real(z) + real(y) * imag(w) - imag(y) * real(w))'
			M42 = '2 * (real(x) * imag(z) - imag(x) * real(z) - real(y) * imag(w) + imag(y) * real(w))'
			M43 = '2 * (real(x) * imag(w) - imag(x) * real(w) + real(y) * imag(z) - imag(y) * real(z))'
			M44 = '2 * (real(x) * real(w) + imag(x) * imag(w) - real(y) * real(z) - imag(y) * imag(z))'

			res = '0.5 * ((' + M41 + ') * a + (' + M42 + ') * b + (' + M43 + ') * c +  (' + M44 + ') * d)'
			return Field(ne.evaluate(res), self.electric_field.grid)
		else:
			# This is a vector field.
			return -2 * np.imag(self.electric_field[0,:] * self.electric_field[1,:].conj())

	@property
	def stokes_vector(self):
		'''The Stokes vector.
		'''
		if self.is_scalar:
			# This is a scalar field and thus we return an unpolarized Stokes vector.
			stokes_vector = Field(np.zeros((4,self.grid.size)), self.grid)
			stokes_vector[0, :] = np.abs(self.electric_field)**2

			return stokes_vector
		elif self.is_partially_polarized:
			# This is a tensor field.
			mueller_matrix = jones_to_mueller(self.electric_field)

			return field_dot(mueller_matrix, self._input_stokes_vector)
		else:
			# This is a vector field and thus we return a fully polarized Stokes vector.
			stokes_vector = Field(np.zeros((4,self.grid.size)), self.grid)

			stokes_vector[0, :] = self.I
			stokes_vector[1, :] = self.Q
			stokes_vector[2, :] = self.U
			stokes_vector[3, :] = self.V

			return stokes_vector

	@property
	def degree_of_polarization(self):
		'''The degree of polarization.
		'''
		return np.sqrt(self.Q**2 + self.U**2 + self.V**2) / self.I

	@property
	def degree_of_linear_polarization(self):
		'''The degree of linear polarization.
		'''
		return np.sqrt(self.Q**2 + self.U**2) / self.I

	@property
	def angle_of_linear_polarization(self):
		'''The angle of linear polarization.
		'''
		return 0.5 * np.arctan2(self.U, self.Q)

	@property
	def degree_of_circular_polarization(self):
		'''The degree of circular polarization.
		'''
		return self.V / self.I

	@property
	def ellipticity(self):
		'''The ratio of the minor to major axis of the electric
		field polarization ellipse.
		'''
		return self.V / (self.I + np.sqrt(self.Q**2 + self.U**2))

	@property
	def is_polarized(self):
		'''If the wavefront can be polarized.
		'''
		return self.electric_field.tensor_order in [1, 2]

	@property
	def is_partially_polarized(self):
		'''If the wavefront can be partially polarized.
		'''
		return self.electric_field.tensor_order == 2

	@property
	def is_scalar(self):
		'''If the wavefront uses the scalar approximation.
		'''
		return self.electric_field.is_scalar_field

	@property
	def intensity(self):
		'''The total intensity of the wavefront as function of 2D position on the plane.
		'''
		return self.I

	@property
	def amplitude(self):
		'''The amplitude of the wavefront as function of 2D position on the plane.
		'''
		return np.abs(self.electric_field)

	@property
	def phase(self):
		'''The phase of the wavefront as function of 2D position on the plane.
		'''
		phase = np.angle(self.electric_field)
		return Field(phase, self.electric_field.grid)

	@property
	def real(self):
		'''The real part of the wavefront as function of 2D position on the plane.
		'''
		return np.real(self.electric_field)

	@property
	def imag(self):
		'''The imaginary part of the wavefront as function of 2D position on the plane.
		'''
		return np.imag(self.electric_field)

	@property
	def power(self):
		'''The power of each pixel in the wavefront.
		'''
		return self.intensity * self.grid.weights

	@property
	def total_power(self):
		'''The total power in this wavefront.
		'''
		return np.sum(self.power)

	@total_power.setter
	def total_power(self, p):
		self.electric_field *= np.sqrt(p / self.total_power)

def jones_to_mueller(jones_matrix):
	'''Convert a Jones matrix to a Mueller matrix.

	Parameters
	----------
	jones_matrix : ndarray or tensor field
		The Jones matrix/matrices to convert to Mueller matrix/matrices.

	Returns
	-------
	ndarray or tensor Field
		The Mueller matrix/matrices.
	'''
	return np.real(field_dot(_U_matrix, field_dot(field_kron(jones_matrix, jones_matrix.conj()), _U_matrix.conj().T)))



from cycler import cycler
from . import colors

# Blue, red, green, orange, purple, brown
palettes = {
	'dark': cycler(color=[colors.blue_700, colors.red_700, colors.green_700, colors.orange_700, colors.purple_700, colors.brown_700])
}

def set_color_scheme(dark=False, publication_quality=False, cmap='viridis'): # pragma: no cover
	"""
	Apply a color scheme to all matplotlib figures. The setting
	publication_quality uses LaTeX for all text in the figure.
	"""
	import warnings
	warnings.warn('set_color_scheme() is deprecated. Copy the color scheme to your own file if you want to continue using it in the future.',
		warnings.DeprecationWarning, stacklevel=2)

	import matplotlib as mpl

	mpl.rc('lines', linewidth=1.5, markeredgewidth=0.25)
	mpl.rc('image', cmap=cmap)
	mpl.rc('legend', scatterpoints=1, numpoints=1, labelspacing=0.3)
	mpl.rc('axes.formatter', limits=(-4,4))
	mpl.rc('text.latex', preamble=['\\usepackage{amsmath}'])

	mpl.rc('xtick', labelsize='small')
	mpl.rc('ytick', labelsize='small')
	mpl.rc('axes', titlesize='medium', labelsize='medium')
	mpl.rc('legend', fontsize='medium')

	mpl.rc('savefig', transparent=True)

	if dark:
		mpl.rc('axes', prop_cycle=palettes['light'], facecolor='k', labelcolor='w', edgecolor='w')
		mpl.rc('xtick', color='w')
		mpl.rc('ytick', color='w')
		mpl.rc('grid', color='w')
		mpl.rc('figure', facecolor='k', edgecolor='k')
		mpl.rc('text', color='w')
	else:
		mpl.rc('axes', prop_cycle=palettes['dark'], facecolor='w', labelcolor='k', edgecolor='k')
		mpl.rc('xtick', color='k')
		mpl.rc('ytick', color='k')
		mpl.rc('grid', color='k')
		mpl.rc('figure', facecolor='w', edgecolor='w')
		mpl.rc('text', color='k')

	if publication_quality:
		mpl.rc('text', usetex=True)
		mpl.rc('font', family='sans-serif')
		mpl.rc('font', serif=['computer modern roman'], monospace=['computer modern typewriter'])
		mpl.rcParams['font.sans-serif'] = ['computer modern sans serif']

		mpl.rc('font', size=11)
		mpl.rc('figure', figsize=(7.2, 5.1))
	else:
		mpl.rc('text', usetex=False)
		mpl.rc('font', family='sans-serif')
		mpl.rc('font', serif=['Bitstream Vera Serif', 'New Century Schoolbook', 'Century Schoolbook L', 'Utopia',
			'ITC Bookman', 'Bookman','Nimbus Roman No9 L', 'Times New Roman', 'Times', 'Palatino', 'Charter', 'serif'])
		mpl.rcParams['font.sans-serif'] = ['Bitstream Vera Sans', 'Lucida Grande', 'Verdana', 'Geneva', 'Lucid', 'Arial',
			'Helvetica', 'Avant Garde', 'sans-serif']
		mpl.rc('font', monospace=['Bitstream Vera Sans Mono', 'Andale Mono', 'Nimbus Mono L', 'Courier New', 'Courier',
			'Fixed', 'Terminal', 'monospace'])

		mpl.rc('figure', figsize=(10, 7.1))
		mpl.rc('font', size=14)
